"""
Module sa_score
=================
This modules contains the functions provided by Ertl et al at:

https://github.com/rdkit/rdkit/blob/master/Contrib/SA_Score/sascorer.py

It was simply edited to suit my needs and tastes, I do not take credit for it.

It uses a pre-computed model to score structures based on the presence or
absence of fragments.
"""
# standard
import math
import gzip
# data handling
import pickle
# chemoinformatics
from rdkit import Chem
from rdkit.Chem import Mol
from rdkit.Chem import rdMolDescriptors


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASSES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class SAScorer:
    """A class for computing the Synthetic Accessibility score, after Ertl et al.
    Neither the algorithm nor the model are mine, I just downloaded them from the ref
    and adapted the code for my project.

    The original model is file is required for scoring molecules, its location
    needs to be provided as argument.

    Reference:
    https://github.com/rdkit/rdkit/blob/master/Contrib/SA_Score/sascorer.py
    """

    def __init__(self, file_model: str):
        self.fscores = self.load_model(file_model)
        self.file_model = file_model

    def __repr__(self):
        return f"SAScorer with model loaded at '{self.file_model}'"

    def load_model(self, file_model: str):
        data = pickle.load(gzip.open(file_model))
        outDict = {}
        for i in data:
            for j in range(1, len(i)):
                outDict[i[j]] = float(i[0])
        return outDict

    def _numBridgeheadsAndSpiro(self, mol, ri=None):
        nSpiro = rdMolDescriptors.CalcNumSpiroAtoms(mol)
        nBridgehead = rdMolDescriptors.CalcNumBridgeheadAtoms(mol)
        return nBridgehead, nSpiro

    def score(self, m: Mol) -> float:
        """Compute the Synthetic Accessiblity score of a molecule.
        It ranges from 1 (very easy to make) to 10 (very hard to make).

        :param mol: the input molecule
        :return: the score
        """
        # fragment score
        fp = rdMolDescriptors.GetMorganFingerprint(m, 2)  # circular fingerprint *radius*
        fps = fp.GetNonzeroElements()
        score1 = 0.
        nf = 0
        for bitId, v in fps.items():
            nf += v
            sfp = bitId
            score1 += self.fscores.get(sfp, -4) * v
        score1 /= nf

        # features score
        nAtoms = m.GetNumAtoms()
        nChiralCenters = len(Chem.FindMolChiralCenters(m, includeUnassigned=True))
        ri = m.GetRingInfo()
        nBridgeheads, nSpiro = self._numBridgeheadsAndSpiro(m, ri)
        nMacrocycles = 0
        for x in ri.AtomRings():
            if len(x) > 8:
                nMacrocycles += 1

        sizePenalty = nAtoms**1.005 - nAtoms
        stereoPenalty = math.log10(nChiralCenters + 1)
        spiroPenalty = math.log10(nSpiro + 1)
        bridgePenalty = math.log10(nBridgeheads + 1)
        macrocyclePenalty = 0.
        # ---------------------------------------
        # This differs from the paper, which defines:
        #  macrocyclePenalty = math.log10(nMacrocycles+1)
        # This form generates better results when 2 or more macrocycles are present
        if nMacrocycles > 0:
            macrocyclePenalty = math.log10(2)

        score2 = 0. - sizePenalty - stereoPenalty - spiroPenalty - bridgePenalty - macrocyclePenalty

        # correction for the fingerprint density
        # not in the original publication, added in version 1.1
        # to make highly symmetrical molecules easier to synthetise
        score3 = 0.
        if nAtoms > len(fps):
            score3 = math.log(float(nAtoms) / len(fps)) * .5

        sascore = score1 + score2 + score3

        # need to transform "raw" value into scale between 1 and 10
        min = -4.0
        max = 2.5
        sascore = 11. - (sascore - min + 1) / (max - min) * 9.
        # smooth the 10-end
        if sascore > 8.:
            sascore = 8. + math.log(sascore + 1. - 9.)
        if sascore > 10.:
            sascore = 10.0
        elif sascore < 1.:
            sascore = 1.0

        return sascore



"""
Module fragment_combination_point
==================================
This modules contains the functions for annotating and parsing fragment
combination point labels in fragments.
"""


# standard
import logging
import re
import string
from collections import defaultdict
# chemoinformatics
from rdkit.Chem import Mol
# docs
from typing import Union
from typing import Tuple
from typing import List


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def idx_to_label(n: int) -> str:
    """Convert a number to a corresponding letter in the alphabet.
    In case the number is higher than the number of letters in the english alphabet, then
    a second character is appended.

    >>> For instance:
    >>>idx_to_label(0)
    >>> 'a'
    >>> idx_to_label(25)
    >>> 'z'
    >>> idx_to_label(26)
    >>> 'aa'

    This function was inspired after:
    https://stackoverflow.com/questions/2267362/how-to-convert-an-integer-to-a-string-in-any-base

    :param n: the input number
    :return: the corresponding string
    """
    alphabet_size = 26
    digits = []
    n += 1
    while n:
        digits.append(int(n % alphabet_size - 1))
        n //= alphabet_size

    digits.reverse()
    return ''.join([string.ascii_lowercase[i] for i in digits])


def find_symmetry_groups(mol: Mol) -> Tuple[Tuple[int]]:
    """Search for symmetry groups within a molecule.
    Symmetry groups are defined as atoms that can have equivalent
    positions in substructure matches.

    This function has been copied from
    https://sourceforge.net/p/rdkit/mailman/message/27897393/

    :param mol: the input molecule
    :return: a tuple of tuples of atom indices. Each tuple represents a symmetry group and inside are the atom indices.
    """
    equivs = defaultdict(set)
    matches = mol.GetSubstructMatches(mol, uniquify=False)
    for match in matches:
        for idx1, idx2 in enumerate(match):
            equivs[idx1].add(idx2)
    classes = set()
    for s in equivs.values():
        classes.add(tuple(s))
    return tuple(classes)


def get_fcp_labels(mol: Mol) -> dict:
    """Search for symmetry groups within the input molecule and defines
    accordingly a label each atom. This label is composed of 2 parts:

        - a prefix: sequential number that gets incremented for each group
        - a suffix: sequential letter or string that gets incremented for each element in group


    This means that equivalent atoms (within the same group) are given the same prefix but different suffixes.

    >>> For instance:
    >>> '1a', '1b', '2'  # 2 equivalent atoms in position 1
    >>> '1', '2', '3', '4', '5', '6'  # a 6-atoms fragment with no symmetry group
    >>> '1a', '1b', '1c', '1d', '1e', '1f'  # a benzene or cyclohexane

    .. note:: labels prefixes begin at 1, not 0 like atom indices.

    :param mol: the input molecule
    :return: a dictionary with the correspondance between atom indices (keys) and atom labels (values)
    """
    symmetry_groups = find_symmetry_groups(mol)
    logging.debug('Symmetry groups: %s', symmetry_groups)
    d_fcp_labels = {}
    for i, aidx_eq in enumerate(symmetry_groups):
        for j, aidx in enumerate(aidx_eq):
            if len(aidx_eq) > 1:
                suffix = idx_to_label(j)
            else:
                suffix = ''
            fcp_label = f"{i+1}{suffix}"
            d_fcp_labels[aidx] = fcp_label

    logging.debug('FCP labels: %s', d_fcp_labels)
    return d_fcp_labels


# def simplify_fcp(fcp: List[str]) -> List[str]:
#     """Remove suffixes from each fcp within provided list.
#
#     >>> For instance:
#     >>> '1a', '1b', '2' becomes: '1', '1', '2'
#
#     :param fcp: a list of Fragment Connection Points
#     :return: the simplified list
#     """
#     return [re.sub("[^0-9]", "", x) for x in fcp]


def clear_fcp_suffixes(fcp_str: str) -> List[str]:
    """Remove suffixes from each fcp within provided string.

    >>> For instance:
    >>> '1a,1b,2' becomes: '1,1,2'

    :param fcp: a list of Fragment Connection Points
    :return: the simplified list
    """
    return re.sub("[^0-9,]", "", fcp_str)


def clear_fcp_suffixes_in_edges(edges: list) -> list:
    """Remove suffixes from each fcp within the attributes found in a list of edges

    :param edges: a list of edges ([[u1, v1, d1], [u2, v2, d2]]), with u and v the nodes and d a dictionary of attributes
    :return: the updated edges
    """
    for e in edges:
        e[2]['fcp_1'] = clear_fcp_suffixes(e[2]['fcp_1'])
        e[2]['fcp_2'] = clear_fcp_suffixes(e[2]['fcp_2'])

    return edges


def count_symmetry_groups(val: Union[Mol, List[str]]) -> int:
    """Return the number of symmetry groups found in a fragment.

    :param val: either a fragment as Mol object or the fragment connection points as dict
    :return: the count of symmetry groups
    """
    # init
    if isinstance(val, Mol):
        val = get_fcp_labels(val)
    elif not isinstance(val, dict):
        raise ValueError(f"Error! Input value is neither a Molecule nor a dictionary: '{val}' ({type(val)})")
    fcp = list(val.values())
    # regroup by symmetry group
    d = {}
    # populate each groups with the equivalent atom indices
    for s in fcp:
        m = re.search(r"[a-z]", s)
        if m:
            idx = m.start()
            prefix = s[:idx]
            if prefix in d.keys():
                d[prefix] += 1
            else:
                d[prefix] = 1

    # count only groups with more than 1 element
    return len([x for x in d.values() if x > 1])



"""
Module load
================

A module containing the Loader class, used for storing DataFrames with molecules
on disk.
"""

# standard
import logging
import gzip
from pathlib import Path
import subprocess
# data science
import psycopg2
import pandas as pd
from pandas import DataFrame
import h5py
# chemoinformatics
from rdkit import Chem
# docs
from typing import List
from typing import Union
# dev
from npfc import utils
import warnings


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GLOBALS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


CONVERTERS = {'molblock': lambda x: Chem.MolFromMolBlock(x),
              'smiles': lambda x: Chem.MolFromSmiles(x),
              'rdkit': lambda x: x,  # nothing to do here, but removes the needs for more if/elif
              }

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def file(input_file: str,
         in_id: str = 'idm',
         in_mol: str = 'mol',
         csv_sep: str = '|',
         mol_format: str = 'rdkit',
         out_id: str = 'idm',
         out_mol: str = 'mol',
         keep_props: bool = True,
         decode: Union[bool, List[str]] = True,
         ) -> DataFrame:
    """Load a file into a DataFrame.

    :param input_file: the input file to load
    :param in_id: the column/property to use for molecule ids
    :param in_mol: the column to use for molecules (irrerlevant for SDF)
    :param csv_sep: the column separator to use for parsing the input file (CSV)
    :param mol_format: the input format for molecules
    :param out_id: the column name used for storing molecule ids
    :param out_mol: the column name used for storing molecules
    :param keep_props: keep all properties found in the input file. If False, then only out_id and out_mol are kept.
    :param decode: decode base64 strings into objects. Columns with encoded objects are labelled with a leading '_'. For molecules, reserved names are 'mol' and 'mol_frag'.
    :return: a DataFrame
    
    ..warning:: if a 'idm' property exists in the input file but the user picks another property for in_id, the pre-existing 'idm' will be renamed into 'idm_2'. 
    """
    # check arguments
    utils.check_arg_input_file(input_file)
    format, compression = utils.get_file_format(input_file)
    logging.debug("Loading file '%s' (format=%s, compression=%s)", input_file, format, compression)

    # read mols
    if format == 'SDF':
        df = _from_sdf(input_file, compression=compression)
        logging.debug("Decode is set to False")
        decode = False
    elif format == 'HDF':
        df = _from_hdf(input_file)
    else:  # has to be CSV
        df = _from_csv(input_file, csv_sep=csv_sep, compression=compression)

    # process data
    
    # in case in_id is different than default but some column with the same name as out_id already exists
    if in_id != out_id and out_id in df.columns:
        df.rename({out_id: f"{out_id}.2"}, axis=1, inplace=True)
        logging.debug('in_id=%s, out_id=%s, but out_id already present. Renaming existing out_id into %s', in_id, out_id, f"{out_id}.2.")

    # out_id
    if out_id is None and in_id is not None:
        out_id = in_id
    elif out_id != in_id:
        df.rename({in_id: out_id}, axis=1, inplace=True)

    # out_mol
    logging.debug('in_mol=%s, out_mol=%s', in_mol, out_mol)
    if out_mol is None and in_mol is not None:
        out_mol = in_mol
    elif out_mol != in_mol and in_mol in df.columns:
        df.rename({in_mol: out_mol}, axis=1, inplace=True)

    # keep_props
    if not keep_props:
        logging.debug("Found properties: %s", ', '.join([c for c in df.columns]))
        logging.debug("Keeping properties: %s", ', '.join([out_id, out_mol]))
        logging.debug("Dropping properties: %s", ', '.join([c for c in df.columns if c not in (out_id, out_mol)]))
        df.drop([c for c in df.columns if c not in (out_id, out_mol)], axis=1, inplace=True)

    # mol_format
    if out_mol in df.columns:
        df[out_mol] = df[out_mol].map(CONVERTERS[mol_format])

    # decode
    if decode:
        # faster way for mols
        for col in ('mol', 'mol_frag', 'mol_frag_1', 'mol_frag_2', 'mol_rdkit'):
            if col in df.columns:
                logging.debug("Decoding column='%s'", col)
                df[col] = df[col].map(utils.decode_mol)

        # other objects are labelled with leading '_'
        for col in df.columns:
            if col.startswith('_') and col != '_Name':
                logging.debug("Decoding column='%s'", col)
                df[col] = df[col].map(utils.decode_object)

    # consistent id comparison accross datasets
    # columns to convert to str
    cols_str = [out_id, 'idf1', 'idf2']
    for col in cols_str:
        if col in df.columns:
            df[col] = df[col].astype(str)

    logging.debug("First 3 rows loaded:\n\n%s\n", df.head(3))

    return df


def _from_sdf(input_sdf: str, col_mol: str = 'mol', compression: str = None):
    if compression is None:
        logging.debug("Read sdf from uncompressed file")
        FH = open(input_sdf, 'rb')
        FH_raw = open(input_sdf, 'rb')
    elif compression == 'gzip':
        logging.debug("Read sdf from compressed file (%s)", compression)
        FH = gzip.open(input_sdf)
        FH_raw = gzip.open(input_sdf)
    else:
        raise ValueError(f"Error! Unknown compression type for SDF: '{compression}'")
    # init
    i = 0
    rows = []
    row_idx = []
    # double iteration over sdf but generators so it should be ok memory-wise
    for mol, mol_raw in zip(Chem.ForwardSDMolSupplier(FH), Chem.ForwardSDMolSupplier(FH_raw, sanitize=False)):
        try:
            # properties
            row = dict((k, mol_raw.GetProp(k)) for k in mol_raw.GetPropNames())
            # molecule title
            row['_Name'] = mol_raw.GetProp('_Name')
            # mol
            row[col_mol] = mol
        except AttributeError:
            logging.warning(f"Molecule #{i} could not be parsed and was skipped!")
            row = None
        # record entry
        if row is not None:
            rows.append(row)
            row_idx.append(i)
        i += 1

    return DataFrame(rows, index=row_idx)


def _from_hdf(input_hdf: str):
    return pd.read_hdf(input_hdf, key=Path(input_hdf).stem)


def _from_csv(input_csv: str, csv_sep: str = '|', compression: str = None):
    if compression is None:
        return pd.read_csv(input_csv, sep=csv_sep)
    elif compression == 'gzip':
        return pd.read_csv(input_csv, sep=csv_sep, compression=compression)


def count_mols(input_file: str,
               keep_uncompressed: bool = False,
               hdf_key: str = None,
               csv_header: bool = True,
               engine: str = 'unix'):
    """Count the number of molecules found in a text file.
    In case the file is compressed (gzip), it is uncompressed first. The resulting
    uncompressed file can be kept for further use.

    No grouping is performed, so duplicate molecules will be counted as well.

    #### this function failed to the ZINC (9,902,598)

    :param input_file: the input file
    :param keep_uncompressed: if the input file is compressed (gzip), do not remove the uncompressed file when finished
    :param hdf_key: specify the key to access data for HDF input. By default, the filename is used. 
    :param csv_header: specifiy whether the CSV input file contains headers or not, so that an extra line is not counted.
    :param engine: either 'unix' or 'python'. By default, UNIX commands (grep for SDF format, wc for CSV for format) are used within Python for maximum efficiency and pure Python code is used as fall-back only. If set to python, then UNIX commands will not be performed.
    """
    utils.check_arg_input_file(input_file)
    format, compression = utils.get_file_format(input_file)
    logging.debug("input_file='%s' (format: %s, compression: %s)", input_file, format, compression)

    # in case of compressed file, uncompress it
    if compression == 'gzip':
        uncompressed_file = input_file.split('.gz')[0]
        with open(uncompressed_file, 'wb') as uncompressed, gzip.open(input_file, 'rb') as compressed:
            bindata = compressed.read()
            uncompressed.write(bindata)
        input_file = uncompressed_file

    # count mols
    if engine == 'unix':
        fall_back = False
    else:
        fall_back = True

    if format == 'SDF':
        # try to use a very fast LINUX command first, and if not available then switch to slower Python code
        if not fall_back:
            try:
                # look for the '$$$$' pattern in the SDF file
                logging.debug("Using UNIX command grep to count molecules in SDF input file.")
                result = subprocess.check_output(f"grep -c '$$$$' {input_file}", shell=True).decode('utf-8').strip()
            except KeyError:
                fall_back = True
                logging.debug("Counting mols with linux grep command failed, switching to pure-Python method.")
        if fall_back:
            logging.debug("Using Python code to count molecules in SDF input file.")
            result = 0
            rows = (row for row in open(input_file, 'r'))
            for row in rows:
                if row.strip() == "$$$$":
                    result += 1

    elif format == 'HDF':
        # no alternative UNIX commands for HDF files, only Python
        if hdf_key is None:
            hdf_key = Path(input_file).stem
        # get the size of a dataframe
        result = h5py.File(input_file, 'r')[hdf_key]['axis1'].shape[0]  # data is not loaded, so very fast and no memory usage

    elif format == 'CSV':
        # try to use a very fast LINUX command first, and if not available then switch to slower Python code
        if not fall_back:
            try:
                # any text-based file with 1 molecule per row (csv, smiles, inchi, etc.)
                result = subprocess.check_output(f"wc -l {input_file}", shell=True).decode('utf-8').split()[0].strip()
            except:
                fall_back = True
                logging.debug("Counting mols with linux wc command failed, switching to pure-Python method.")
        if fall_back:
            logging.debug("Using UNIX command wc to count molecules in CSV input file.")
            result = 0
            rows = (row for row in open(input_file))
            for row in rows:
                result += 1
    else:
        raise ValueError(f"Error! Input file format is not valid ('{input_file}' is neither SDF, HDF or CSV, with or without .gz suffix).")

    logging.debug('FALL_BACK: %s', fall_back)

    # read count
    count = int(result)
    # for CSV only, substract the header line that was counted in the beginning of the file
    if format == 'CSV' and csv_header:
        count -= 1

    # remove uncompressed file
    if compression and not keep_uncompressed:
        Path(uncompressed_file).unlink()

    return count



"""
Module report
================

A module with helper functions for computing pre-defined plots for the analysis
of fragment combinations.
"""

# standard
import warnings
import logging
import argparse
import sys
from shutil import rmtree
from datetime import datetime
import re
from pathlib import Path
from collections import OrderedDict
# data handling
import numpy as np
import json
import pandas as pd
from pandas import DataFrame
import networkx as nx
# data visualization
from matplotlib import pyplot as plt
import seaborn as sns
import matplotlib.ticker as mticker
# from pylab import savefig
from adjustText import adjust_text
# chemoinformatics
import rdkit
from rdkit import Chem
from rdkit.Chem import Mol
# docs
from typing import List
from typing import Tuple
from rdkit import RDLogger
# custom libraries
import npfc
from npfc import utils
from npfc import load
from npfc import save
from npfc import draw
from npfc import fragment_combination
from multiprocessing import Pool


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GLOBALS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


DEFAULT_PALETTE = {'gray': '#808080',
                   'green': '#2CA02C',
                   'blue': '#378EBF',
                   'red': '#EB5763',
                   'orange': '#EBC81A',
                   }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def init_report_globals():
    """This function is used to init a reporting session by setting the default
    parameters for displaying information.
    """
    pd.set_option('display.max_columns', 20)
    pd.set_option('display.max_rows', 20)
    pd.set_option('max_colwidth', 70)


def print_title(title: str, level: int = 2, pad: int = 160):
    """Print a title with padding.

    :param title: title to print
    :param level: level of importance of the title (lower level is more important, min=1)
    :param pad: a padding added for highlighting the title ("="*pad)
    """
    if level == 1:
        # super title
        logging.info("=" * pad)
        logging.info(title.upper().center(pad))
        logging.info("=" * pad)
    elif level == 2:
        # normal title
        logging.info("*" * pad)
        logging.info(title.upper().center(pad))
        logging.info("*" * pad)
    elif level == 3:
        # subtitle
        logging.info("-" * pad)
        logging.info(title.upper().center(pad))
        logging.info("-" * pad)
    elif level == 4:
        # sub-subtitle
        logging.info("." * pad)
        logging.info(title.upper().center(pad))
        logging.info("." * pad)
    else:
        raise ValueError(f"ERROR! UNKNOWN LEVEL! ({level})")


def _get_chunks(WD: str, pattern: str) -> List[str]:
    """This function returns the list of all chunks found in a given directory.
    It does not use the _000_ notation because it can also be applied to fragments.

    :param WD: the WD where all subfolders for each step are present (i.e. 'natural/dnp/data')
    :param pattern: the pattern that desired chunks contain (i.e. '1_input/data/*([0-9][0-9][0-9])?.sdf.gz')
    :return: the list of chunks
    """
    WD = Path(WD)
    pattern = re.compile(pattern)
    chunks = [str(x) for x in list(WD.glob("*"))]
    return sorted(list(filter(pattern.match, chunks)))





def save_barplot(df: DataFrame,
                 output_plot: str,
                 x_name: str,
                 y_name: str,
                 title: str,
                 color: str,
                 x_label: str = None,
                 y_label: str = None,
                 rotate_x: int = 0,
                 perc_labels: str = None,
                 perc_label_size: int = 15,
                 fig_size: Tuple[int] = (24, 12),
                 force_order: bool = True,
                 print_rank: bool = False,
                 ):
    """This function helps for computing automated barplots using seaborn.
    It sets up somewhat standardized figure output for a harmonized rendering.

    :param df: the DataFrame with data to plot
    :param output_plot: the output plot full file name
    :param x_name: DF column name to use for x-axis
    :param y_name: DF column name to use for y-axis
    :param x_label: the name to display on the plot for x-axis
    :param y_label: the name to display on the plot for y-axis
    :param rotate_x: rotate the x-axis ticks anticlock-wise
    :param perc_labels: DF column name to use display percentage labels above bars
    :param perc_label_size: annotation text size for perc_labels
    :param color: color to use for bars, theoritically could also be a list of colors
    :param fig_size: tuple of integers defining the plot dimensions (x, y)
    :param force_order: force the plot to display the bars in the provided order. I do not know why sometimes this is needed, sometimes not depending on the plot.
    :param print_rank: display the rank as #i above the bar label
    :return: the figure in searborn format
    """
    # detect format from file extension
    format = Path(output_plot).suffix[1:].lower()
    if format != 'svg' and format != 'png':
        raise ValueError(f"ERROR! UNKNOWN PLOT FORMAT! ('{format}')")
    logging.debug(f"FORMAT FOR PLOT: '{format}'")

    # delete existing file for preventing stacking of plots
    p = Path(output_plot)
    if p.exists():
        p.unlink()

    # general style for plotting
    # sns.set(rc={'figure.figsize': fig_size})
    plt.figure(figsize=fig_size)
    sns.set_style('whitegrid', {'axes.edgecolor': '0.2'})
    sns.set_context("paper", font_scale=2)

    # barplot
    if force_order:
        ax = sns.barplot(x=df[x_name], y=df[y_name], color=color, order=df[x_name])
    else:
        ax = sns.barplot(x=df[x_name], y=df[y_name], color=color)
    ax.set_title(title, fontsize=24, y=1.02)
    ax.tick_params(labelsize=20)
    ax.tick_params(axis='x', rotation=rotate_x)
    # ax.set_xticklabels(df[x_name])
    ax.set_xlabel(x_label, fontsize=25, labelpad=20)
    ax.set_ylabel(y_label, fontsize=25, labelpad=20)
    # if no entry, then y ticks get all confused
    if df[y_name].sum() == 0:
        ax.set_yticks((0, 1, 2, 3, 4, 5))
        ax.set_ylim((0, 5))

    # format y labels
    label_format = '{:,.0f}'
    ticks_loc = ax.get_yticks().tolist()
    ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_yticklabels([label_format.format(x) for x in ticks_loc])

    # add percentage annotations
    if perc_labels is not None:
        for a, i in zip(ax.patches, range(len(df.index))):
            row = df.iloc[i]
            ax.text(row.name, a.get_height(), row[perc_labels], color='black', ha='center', fontdict={'fontsize': perc_label_size})

    # save
    figure = ax.get_figure()
    figure.subplots_adjust(bottom=0.2)
    figure.savefig(output_plot, format=format, dpi=600)
    plt.clf()
    plt.close()

    return figure


def save_lineplot(df: DataFrame,
                  output_plot: str,
                  x_name: str,
                  y_name: str,
                  title: str,
                  color: str,
                  x_label: str = None,
                  y_label: str = None,
                  rotate_x: int = 0,
                  perc_labels: str = None,
                  perc_label_size: int = 15,
                  fig_size: Tuple[int] = (24, 12),
                  fill: bool = False,
                  ):
    """This function helps for computing automated lineplot using seaborn.
    It sets up somewhat standardized figure output for a harmonized rendering.

    :param df: the DataFrame with data to plot
    :param output_plot: the output plot full file name
    :param x_name: DF column name to use for x-axis
    :param y_name: DF column name to use for y-axis
    :param x_label: the name to display on the plot for x-axis
    :param y_label: the name to display on the plot for y-axis
    :param rotate_x: rotate the x-axis ticks anticlock-wise
    :param perc_labels: DF column name to use display percentage labels above bars
    :param perc_label_size: annotation text size for perc_labels
    :param color: color to use for bars, theoritically could also be a list of colors
    :param fig_size: tuple of integers defining the plot dimensions (x, y)
    :return: the figure in searborn format
    """
    # detect format from file extension
    format = Path(output_plot).suffix[1:].lower()
    if format != 'svg' and format != 'png':
        raise ValueError(f"ERROR! UNKNOWN PLOT FORMAT! ('{format}')")
    logging.debug(f"FORMAT FOR PLOT: '{format}'")

    # delete existing file for preventing stacking of plots
    p = Path(output_plot)
    if p.exists():
        p.unlink()

    # general style for plotting
    # sns.set(rc={'figure.figsize': fig_size})
    plt.figure(figsize=fig_size)
    sns.set_style('whitegrid', {'axes.edgecolor': '0.2'})
    sns.set_context("paper", font_scale=2)

    # lineplot
    fig, ax = plt.subplots(figsize=fig_size)
    ax = sns.scatterplot(x=x_name, y=y_name, data=df, ax=ax, s=20)
    ax = sns.lineplot(x=df[x_name], y=df[y_name], color=color, ax=ax)
    # fix xticks that would contain floats instead of ints
    plt.xticks(np.arange(min(df[x_name]), max(df[x_name]) + 1, 1))  # https://stackoverflow.com/questions/30327153/seaborn-pylab-changing-xticks-from-float-to-int
    # update all lines width
    for l in ax.lines:
        plt.setp(l, linewidth=3)
    ax.margins(0, 0)
    ax.set_title(title, fontsize=24, y=1.02)
    ax.tick_params(labelsize=20)
    ax.tick_params(axis='x', rotation=rotate_x)
    # ax.set_xticklabels(df[x_name])
    ax.set_xlabel(x_label, fontsize=25, labelpad=20)
    ax.set_ylabel(y_label, fontsize=25, labelpad=20)
    # if no entry, then y ticks get all confused
    if df[y_name].sum() == 0:
        ax.set_yticks((0, 1, 2, 3, 4, 5))
        ax.set_ylim((0, 5))

    # format y labels
    label_format = '{:,.0f}'
    ticks_loc = ax.get_yticks().tolist()
    ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_yticklabels([label_format.format(x) for x in ticks_loc])

    # add percentage annotations
    if perc_labels is not None:
        for i in range(len(df.index)):
            row = df.iloc[i]
            ax.text(row[x_name], row[y_name], row[perc_labels], color='black', ha='center', fontdict={'fontsize': perc_label_size})

    if fill:
        plt.fill_between(df[x_name].values, df[y_name].values)

    # adjust overlapping text annotations (percentages with axes, etc.)
    texts = [x for x in ax.texts]
    adjust_text(texts)

    # save
    figure = ax.get_figure()
    figure.subplots_adjust(bottom=0.2)
    figure.savefig(output_plot, format=format, dpi=600)
    plt.clf()
    plt.close()

    return figure


def save_kdeplot(df: DataFrame,
                 output_plot: str,
                 x_name: str,
                 title: str,
                 color: str,
                 x_label: str = None,
                 y_label: str = None,
                 normalize_x: bool = True,
                 fig_size: Tuple[int] = (24, 12),
                 ):
    """This function helps for computing automated kdeplots using seaborn.
    It sets up somewhat standardized figure output for a harmonized rendering.

    :param df: the DataFrame with data to plot
    :param output_plot: the output plot full file name
    :param x_name: DF column name to use for x-axis
    :param x_label: the name to display on the plot for x-axis
    :param y_label: the name to display on the plot for y-axis
    :param color: color to use for bars, theoritically could also be a list of colors
    :param fig_size: tuple of integers defining the plot dimensions (x, y)
    :return: the figure in searborn format
    """
    # detect format from file extension
    format = Path(output_plot).suffix[1:].lower()
    if format != 'svg' and format != 'png':
        raise ValueError(f"ERROR! UNKNOWN PLOT FORMAT! ('{format}')")
    logging.debug(f"FORMAT FOR PLOT: '{format}'")

    # delete existing file for preventing stacking of plots
    p = Path(output_plot)
    if p.exists():
        p.unlink()

    # general style for plotting
    sns.set(rc={'figure.figsize': fig_size})
    sns.set_style('whitegrid', {'axes.edgecolor': '0.2'})
    sns.set_context("paper", font_scale=2)

    ax = sns.kdeplot(df[x_name], shade=True, label='', color=color)
    ax.set_title(title, fontsize=24, y=1.02)
    ax.tick_params(labelsize=20)
    ax.tick_params(axis='x', rotation=0)
    ax.set_xlim(0, 1)
    # ax.set_xticklabels(df[x_name])
    label_format = '{:,.0%}'
    ticks_loc = ax.get_xticks().tolist()
    ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_xticklabels([label_format.format(x) for x in ticks_loc])

    #ax.set_xticklabels(['{:,.0%}'.format(x) for x in ax.get_xticks()])
    ax.set_xlabel(x_label, fontsize=25, labelpad=20)
    ax.set_ylabel(y_label, fontsize=25, labelpad=20)

    # save
    figure = ax.get_figure()
    figure.savefig(output_plot, dpi=600)
    plt.clf()
    plt.close()

    return figure


def save_distplot(df: DataFrame,
                  output_plot: str,
                  x_name: str,
                  title: str,
                  color: str,
                  x_label: str = None,
                  y_label: str = None,
                  normalize_x: bool = True,
                  fig_size: Tuple[int] = (24, 12),
                  bins=None
                  ):
    """This function helps for computing automated distplots using seaborn.
    It sets up somewhat standardized figure output for a harmonized rendering.

    :param df: the DataFrame with data to plot
    :param output_plot: the output plot full file name
    :param x_name: DF column name to use for x-axis
    :param x_label: the name to display on the plot for x-axis
    :param y_label: the name to display on the plot for y-axis
    :param color: color to use for bars, theoritically could also be a list of colors
    :param fig_size: tuple of integers defining the plot dimensions (x, y)
    :param bins: number of bins
    :return: the figure in searborn format
    """
    # detect format from file extension
    format = Path(output_plot).suffix[1:].lower()
    if format != 'svg' and format != 'png':
        raise ValueError(f"ERROR! UNKNOWN PLOT FORMAT! ('{format}')")
    logging.debug(f"FORMAT FOR PLOT: '{format}'")

    # delete existing file for preventing stacking of plots
    p = Path(output_plot)
    if p.exists():
        p.unlink()

    # general style for plotting
    sns.set(rc={'figure.figsize': fig_size})
    sns.set_style('whitegrid', {'axes.edgecolor': '0.2'})
    sns.set_context("paper", font_scale=2)
    vals = []
    for i in range(len(df)):
        row = df.iloc[i]
        vals += [row[x_name]] * row['Count']
    ax = sns.distplot(vals, kde=False, label='', color=color, bins=bins, hist_kws=dict(alpha=1))
    # ax.set_xlim(0, df[x_name].max())
    ax.set_title(title, fontsize=24, y=1.02)
    ax.tick_params(labelsize=20)
    ax.tick_params(axis='x', rotation=0)

    label_format = '{:,}'
    ticks_loc = ax.get_yticks().tolist()
    ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    ax.set_yticklabels([label_format.format(x) for x in ticks_loc])

    # ax.set_yticklabels(['{:,}'.format(int(x)) for x in ax.get_yticks()])
    ax.set_xlabel(x_label, fontsize=25, labelpad=20)
    ax.set_ylabel(y_label, fontsize=25, labelpad=20)

    # save
    figure = ax.get_figure()
    figure.savefig(output_plot, dpi=600)
    plt.clf()
    plt.close()

    return figure

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASSES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #



"""
Module save
================

A module containing the Saver class, used for storing DataFrames with molecules
on disk.
"""

# standard
import logging
import gzip
import shutil
import os
import numpy as np
import time
from pathlib import Path
from random import random
# data science
import pandas as pd
from pandas import DataFrame
from pandas import HDFStore
# chemoinformatics
from rdkit import Chem
from rdkit.Chem import SDWriter
# docs
from typing import List
# dev
from npfc import utils


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def file(df: pd.DataFrame,
         output_file: str,
         shuffle: bool = False,
         random_seed: int = None,
         chunk_size: int = None,
         encode: bool = True,
         col_mol: str = 'mol',
         col_id: str = 'idm',
         csv_sep: str = '|'):
    """A method for saving DataFrames with molecules to different file types.
    This is handy way of using the Saver class without having to keep a Saver object.

    :param df: the input DataFrame
    :param output_file: the output file
    :param shuffle: randomize records
    :param random_seed: a number for reproducing the shuffling
    :param chunk_size: the maximum number of records per chunk. If this value is unset, no chunking is performed, otherwise each chunk filename gets appended with a suffix: file_XXX.ext.
    :param encode: encode RDKit Mol objects and other objects in predefined columns as base64 strings.
    :param col_mol: if molecules need to be encoded, then the encoding is perfomed on this column.
    :param csv_sep: separator to use in case of csv output
    :return: the list of output files with their number of records
    """
    # check some arguments
    utils.check_arg_output_file(output_file)
    utils.check_arg_bool(shuffle)
    utils.check_arg_bool(encode)

    logging.debug("Excerpt of the data as provided to file function:\n\n%s\n", df.head(5))

    # init
    path_output_file = Path(output_file)
    ext_output_file = path_output_file.suffixes
    output_dir = path_output_file.resolve().parent
    output_files = []
    # for sdf, molecules cannot be encoded
    if ext_output_file[0] == '.sdf' and encode:
        logging.warning(f"Format is SDF, so column '{col_mol}' is not encoded.")
    # avoid pandas warnings
    df = df.copy()
    # shuffle
    if shuffle:
        logging.debug('Shuffling rows before saving file')
        df = df.sample(frac=1, random_state=random_seed)
    # encode predefined data
    # if nothing to encode, just don't
    if len(df.index) == 0:
        logging.warning("DataFrame is empty, skip encoding.")
    # in case there is stuff to encode, encode it:
    elif encode:
        # for SDF files, RDKit Mol objects to use for MolBlocks should not be encoded
        if ext_output_file[0] != '.sdf' and col_mol in df.columns:
            # df[col_mol] = df[col_mol].map(utils.encode_mol_smiles)
            df[col_mol] = df[col_mol].map(utils.encode_mol)
        # other RDKit Mol objects can be encoded though
        for col in ("mol", "mol_frag", "mol_frag_1", "mol_frag_2", "mol_rdkit"):
            if col in df.columns and col != col_mol:
                # df[col] = df[col].map(utils.encode_mol_smiles)
                df[col] = df[col].map(utils.encode_mol)
        # other objects are labelled with leading '_'
        for col in df.columns:
            if col.startswith('_') and col != '_Name':
                df[col] = df[col].map(utils.encode_object)

    logging.debug("Excerpt of the data to save before chuking:\n\n%s\n", df.head(3))

    # chunking
    if chunk_size is None:
        # single output
        _save(df=df, output_file=output_file, col_mol=col_mol, col_id=col_id, suffixes=ext_output_file, key=path_output_file.stem.split('.')[0], csv_sep=csv_sep)
        output_files.append([output_file, len(df.index)])
    else:
        # chunks
        start = 0
        j = 1
        for start in range(0, len(df.index), chunk_size):
            end = start + chunk_size
            output_chunk = str(output_dir) + "/" + path_output_file.stem.split('.')[0] + "_" + str(j).zfill(3) + ''.join(ext_output_file)  # stem returns file.csv for file.csv.gz
            _save(df=df.iloc[start:end], output_file=output_chunk, col_mol=col_mol, col_id=col_id, suffixes=ext_output_file, key=path_output_file.stem.split('.')[0], csv_sep=csv_sep)
            output_files.append([output_chunk, len(df.iloc[start:end].index)])
            j += 1
        logging.debug("%s chunks were created", len(output_files))

    return output_files


def _save(df: DataFrame,
          output_file: str,
          col_mol: str,
          col_id: str,
          suffixes: List[str],
          key: str,
          csv_sep: str):
    """Helper function for the save method.
    Does the actual export to the output file and picks a format based on provided infos.

    :param df: the input DataFrame
    :param suffixes: the suffixes of the output file
    :param key: the key for a HDF file
    :param csv_sep: the separator for a CSV file
    """
    # infer from pandas.to_csv does not work as expected (no compression!)
    # so I need to specify the compression type manually.
    utils.check_arg_output_file(output_file)
    out_format, out_compression = utils.get_file_format(output_file)
    if out_format == 'CSV':
        if out_compression == 'gzip':
            df.to_csv(output_file, sep=csv_sep, compression=out_compression, index=False)
        else:
            df.to_csv(output_file, sep=csv_sep, index=False)
    elif out_format == 'HDF':
        df.to_hdf(output_file, key=key)
    elif out_format == 'SDF':
        # write the uncompressed file
        if out_compression == 'gzip':
            # init
            output_file_base = '.'.join(output_file.split('.')[:-1])
            logging.debug("Output_file_base: %s", output_file_base)
            # write the file uncompressed
            write_sdf(df, output_file_base, molColName=col_mol, idName=col_id, properties=list(df.columns))
            # compress the file
            with open(output_file_base, 'rb') as OUTPUT:
                with gzip.open(output_file, 'wb') as ARCHIVE:
                    shutil.copyfileobj(OUTPUT, ARCHIVE)
                # delete the uncompressed file as it is only a byproduct
                Path(output_file_base).unlink()
        else:
            write_sdf(df, output_file, molColName=col_mol, idName=col_id, properties=list(df.columns))
    elif out_format == 'FEATHER':
        df.to_feather(output_file)
    else:
        raise ValueError(f"Error! Cannot save DataFrame to unexpected format '{suffixes[0]}'.")
    logging.debug("Saved %s records at '%s'.", len(df.index), output_file)


def write_sdf(df, out, molColName='ROMol', idName=None, properties=None, allNumeric=False):
    """
    Redefinition of PandasTools.WriteSDF because RDKit 2019.03.1 is incompatible with Pandas 25.1.

    Write an SD file for the molecules in the dataframe. Dataframe columns can be exported as
    SDF tags if specified in the "properties" list. "properties=list(df.columns)" would export
    all columns.
    The "allNumeric" flag allows to automatically include all numeric columns in the output.
    User has to make sure that correct data type is assigned to column.
    "idName" can be used to select a column to serve as molecule title. It can be set to
    "RowID" to use the dataframe row key as title.
    """
    close = None
    if isinstance(out, str):
        if out.lower()[-3:] == ".gz":
            out = gzip.open(out, "wt")
            close = out.close

    writer = SDWriter(out)
    if properties is None:
        properties = []
    else:
        properties = list(properties)
    if allNumeric:
        properties.extend([
          dt for dt in df.dtypes.keys()
          if (np.issubdtype(df.dtypes[dt], np.floating) or np.issubdtype(df.dtypes[dt], np.integer))
        ])

    if molColName in properties:
        properties.remove(molColName)
    # if idName in properties:
    #     properties.remove(idName)
    writer.SetProps(properties)
    for row in df.iterrows():
        # make a local copy I can modify
        mol = Chem.Mol(row[1][molColName])

        if idName is not None:
            if idName == 'RowID':
                mol.SetProp('_Name', str(row[0]))
            else:
                mol.SetProp('_Name', str(row[1][idName]))
        for p in properties:
            cell_value = row[1][p]
            # Make sure float does not get formatted in E notation
            if np.issubdtype(type(cell_value), np.floating):
                s = '{:f}'.format(cell_value).rstrip("0")  # "f" will show 7.0 as 7.00000
                if s[-1] == ".":
                    s += "0"  # put the "0" back on if it's something like "7."
                mol.SetProp(p, s)
            else:
                mol.SetProp(p, str(cell_value))
        writer.write(mol)
    writer.close()
    if close is not None:
        close()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASSES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class SafeHDF5Store(HDFStore):
    """Implement safe HDFStore by obtaining file lock. Multiple writes will queue if lock is not obtained.

    Edited after:
    https://stackoverflow.com/questions/41231678/obtaining-a-exclusive-lock-when-writing-to-an-hdf5-file
    """

    def __init__(self, *args, **kwargs):
        """Initialize and obtain file lock."""

        interval = kwargs.pop('probe_interval', random())
        self._lock = f"{args[0]}.lock"
        while True:
            try:
                self._flock = os.open(self._lock, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
                break
            except (IOError, OSError):
                time.sleep(interval)

        HDFStore.__init__(self, *args, **kwargs)

    def __exit__(self, *args, **kwargs):
        """Exit and remove file lock."""

        HDFStore.__exit__(self, *args, **kwargs)
        os.close(self._flock)
        os.remove(self._lock)



"""
Module filter
==============
This modules contains the class Filter, which is used to filter molecules using
molecular descriptors.
"""

# data handling
import logging
import re
# chemoinformatics
from rdkit.Chem import Mol
from rdkit.Chem import Crippen
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors
# docs
from typing import List


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def count_violations_lipinski(molecular_weight, slogp, num_hbd, num_hba):
    """Lipinski, J Pharmacol Toxicol Methods. 2000 Jul-Aug;44(1):235-49.
    """
    n = 0
    if molecular_weight < 150 or molecular_weight > 500:
        n += 1
    if slogp > 5:
        n += 1
    if num_hbd > 5:
        n += 1
    if num_hba > 10:
        n += 1
    return n


def count_violations_veber(num_rotatable_bonds, tpsa):
    """Veber DF, Johnson SR, Cheng HY, Smith BR, Ward KW, Kopple KD (June 2002).
    "Molecular properties that influence the oral bioavailability of drug candidates".
    J. Med. Chem. 45 (12): 2615–23.
    """
    n = 0
    if num_rotatable_bonds > 10:
        n += 1
    if tpsa > 140:
        n += 1
    return n


def count_violations_lead_like(molecular_weight, slogp, num_rotatable_bonds):
    """http://zinc.docking.org/browse/subsets/
    Teague, Davis, Leeson, Oprea, Angew Chem Int Ed Engl. 1999 Dec 16;38(24):3743-3748.

    """
    n = 0
    if molecular_weight < 250 or molecular_weight > 350:
        n += 1
    if slogp > 3.5:
        n += 1
    if num_rotatable_bonds > 7:
        n += 1
    return n


def count_violations_ppi_like(molecular_weight, slogp, num_hba, num_rings):
    """Hamon, V., Bourgeas, R., Ducrot, P., Theret, I., Xuereb, L., Basse, M.J., Brunel, J.M., Combes, S., Morelli, X., Roche, P., 2013.
    2P2IHUNTER: a tool for filtering orthosteric protein–protein interaction modulators via a dedicated support vector machine.
    Journal of The Royal Society Interface 11. doi:10.1098/rsif.2013.0860
    """
    n = 0
    if molecular_weight > 400:
        n += 1
    if slogp < 4:
        n += 1
    if num_hba < 4:
        n += 1
    if num_rings < 4:
        n += 1
    return n


def count_violations_fragment_like(molecular_weight, slogp, num_hba, num_hbd):
    """Congreve, M., Carr, R., Murray, C., Jhoti, H., 2003. A “Rule of Three” for fragment-based lead discovery?
    Drug Discovery Today 8, 876–877. doi:10.1016/S1359-6446(03)02831-9
    """
    n = 0
    if molecular_weight >= 300:
        n += 1
    if slogp > 3:
        n += 1
    if num_hba > 3:
        n += 1
    if num_hbd > 3:
        n += 1
    return n


def count_violations_fragment_like_ext(num_fragment_like_violations, tpsa, num_rotatable_bonds):
    """Congreve, M., Carr, R., Murray, C., Jhoti, H., 2003.
    A “Rule of Three” for fragment-based lead discovery? Drug Discovery Today 8, 876–877.
    doi:10.1016/S1359-6446(03)02831-9
    """
    n = num_fragment_like_violations
    if tpsa > 60:
        n += 1
    if num_rotatable_bonds:
        n += 1
    return n


def get_min_max_ring_sizes(mol):
    """Return a tuple wih (minimum, maximum) ring sizes of the input molecule.
    In case the molecule is linear, (0, 0) is returned.
    """
    ring_sizes = [len(x) for x in mol.GetRingInfo().AtomRings()]
    if len(ring_sizes) > 0:
        min_ring_size = min(ring_sizes)
        max_ring_size = max(ring_sizes)
    else:
        min_ring_size = 0
        max_ring_size = 0

    return (min_ring_size, max_ring_size)


DESCRIPTORS = {
              # classical molecular descriptors
              'num_heavy_atoms': lambda x: x.GetNumAtoms(),
              'molecular_weight': lambda x: round(Descriptors.ExactMolWt(x), 4),
              'num_rings': lambda x: rdMolDescriptors.CalcNumRings(x),
              'num_rings_arom': lambda x: rdMolDescriptors.CalcNumAromaticRings(x),
              'elements': lambda x: set([a.GetSymbol() for a in x.GetAtoms()]),
              'molecular_formula': lambda x: rdMolDescriptors.CalcMolFormula(x),
              'num_hbd': lambda x: rdMolDescriptors.CalcNumLipinskiHBD(x),
              'num_hba': lambda x: rdMolDescriptors.CalcNumLipinskiHBA(x),
              'slogp': lambda x: round(Crippen.MolLogP(x), 4),
              'tpsa': lambda x: round(rdMolDescriptors.CalcTPSA(x), 4),
              'num_rotatable_bonds': lambda x: rdMolDescriptors.CalcNumRotatableBonds(x),
              'num_atoms_oxygen': lambda x: len([a for a in x.GetAtoms() if a.GetAtomicNum() == 8]),
              'num_atoms_nitrogen': lambda x: len([a for a in x.GetAtoms() if a.GetAtomicNum() == 7]),
              'fsp3': lambda x: rdMolDescriptors.CalcFractionCSP3(x),
              # custom molecular descriptors
              # ring_sizes:
              # it would have been faster to access only once RingInfo for both min and max,
              # but this is tricky because I would have to start making exceptions in the way
              # the functions are accessed or more complicated downstream process.
              # Indeed, I do not think it is possible to set two dict keys at once
              # within a dict comprehension and it is not that bad for
              # performance to call it twice anyway.
              'ring_size_min': lambda x: min([len(y) for y in x.GetRingInfo().AtomRings()]),
              'ring_size_max': lambda x: max([len(y) for y in x.GetRingInfo().AtomRings()]),
              }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASSES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class Filter:
    """A class for filtering molecules based on molecular descriptors."""

    def __init__(self, descriptors: list = DESCRIPTORS):
        """Create a Filter object."""

        self.descriptors = descriptors

    def compute_descriptors(self, mol: Mol, descriptors: List = None) -> dict:
        """Compute descriptors. A subset of descriptors can be computed if a list
        of descriptor names is provided. To get an idea of what descriptors can be computed,
        the method get_possible_descriptors can be used.

        :param mol: the input molecule
        :param descriptors: the list of descriptors to compute. If none is provided, all possible descriptors are computed.
        :return: a dictionary with all descriptors
        """
        # if no descriptor is specified, compute them all
        if descriptors is None:
            descriptors = list(self.descriptors.keys())

        if len(descriptors) == 0:
            raise ValueError('Error! No descriptor is specified for computation!')

        return {descriptors[i]: self.descriptors[descriptors[i]](mol) for i in range(len(descriptors))}

    def get_possible_descriptors(self) -> List:
        """Return a list of all descriptors that can be computed using this module.

        :return: the list of descriptors that can be computed
        """
        return sorted(list((self.descriptors.keys())))

    def filter_mol(self, mol: Mol, expr: str) -> bool:
        """Filter a molecule based on an expression.
        Two types of expressions are currently supported:

            - inclusion/exclusion
                - 'elements not in C, N, O'
                - 'elements in C, N, O'

            - numeric
                - 'num_heavy_atoms > 3'
                - '100.0 < molecular_weight <= 1000.0'
                - 'num_rings' != 0'
                - 'num_rings == 0'

        :param mol: the input molecule
        :param expr: the filter to apply
        :return: True if the molecule passes the filter, False otherwise
        """
        # init
        split_expr = [s.lower() for s in expr.lower().split()]
        # filters of type: 'elements in C, N, O'
        if 'in' in split_expr:  # 'in' or 'not in'
            return self._eval_set_expr(mol, expr)
        # filters of type: 'num_heavy_atoms > 3'
        return self._eval_numeric_expr(mol, expr.lower())

    def _eval_numeric_expr(self, mol, expr):
        """
        Evaluate if the statements stored in the expression are True or False.
        For now statement is composed of either 3 elements (['molweiht', '<=', '1000'])
        or 5 elements: (['0', '<=', 'molecular_weight', '<=', '1000']).
        ### No check has been added on this number because there might be an expanded functionality
        later on (combining statements with ';'?).
        Descriptors used for the comparisons need to be provided as a dictionary (name: value).

        Possible values for how: numeric, set or literal.
        """
        mol = Mol(mol)
        expr = expr.replace(" ", "")
        split_expr = self._split_expr(expr)  # something like 'molecular_weight', '<=', '1000'
        # replace descriptor names by their values
        split_expr = [self.descriptors[k](mol) if k in self.descriptors.keys() else k for k in split_expr]  # now it is '250.0', '<=', '1000'
        logging.debug("Applying numeric filter: %s", ' '.join(str(v) for v in split_expr))
        # convert all values extracted as string into their type
        split_expr = [float(x) if x not in split_expr[1::2] else x for x in split_expr]  # and now it is 250.0, '<=', 1000.0
        # operators are always at odd positions, whereas values are at even positions
        # and there is always a value on the left and on the right of an operator
        for i in range(1, len(split_expr), 2):
            operator = split_expr[i]
            left = split_expr[i-1]
            right = float(split_expr[i+1])
            if operator == "<=":
                if not left <= right:
                    return False
            elif operator == "<":
                if not left < right:
                    return False
            elif operator == "==":
                if not left == right:
                    return False
            elif operator == "!=":
                if not left != right:
                    return False
            elif operator == ">=":
                if not left >= right:
                    return False
            elif operator == ">":
                if not left > right:
                    return False
        return True

    def _eval_set_expr(self, mol, expr):
        """Helper function for _eval_expr.
        Look for keywords ' in ' and ' not in ' in expression and check the condition
        by defining left as the descriptor and right as the values, i.e.:
        descriptor in values ('elements in H, C, N, O')
        """
        for op in [' not in ', ' in ']:
            pattern = re.compile(op)  # raw string
            hits = [(m.start(0), m.end(0)) for m in re.finditer(pattern, expr)]
            if len(hits) > 0:
                break  # leave asap with op still set to the correct operator
        # in case we did not find anything, just stop
        if len(hits) == 0:
            raise ValueError(f"expected ' not in ' or ' in ' in expr ({expr})")
        expr_split = [e.replace(" ", "") for e in expr.split(op)]
        descriptor = self.descriptors[expr_split[0]](mol)  # left
        values = set(expr_split[1].split(","))  # right
        logging.debug("Applying inclusion/exclusion filter: %s", ''.join(str(v) for v in [descriptor, op, values]))
        if (op == ' in ' and descriptor.issubset(values)) or (op == ' not in ' and not descriptor.issubset(values)):
            return True
        else:
            return False

    def _split_expr(self, expr):
        """Helper function for _eval_expr.
        From a string containing an expression (i.e. 'molecular_weight < 1000'), return
        a list of values and operators (['molecular_weight', '<', '1000']).
        """
        opidx_eq = self._find_opidx("==", expr)
        opidx_diff = self._find_opidx("!=", expr)
        opidx_supeq = self._find_opidx(">=", expr)
        opidx_infeq = self._find_opidx("<=", expr)
        opidx_sup = self._find_opidx(">", expr)
        opidx_inf = self._find_opidx("<", expr)

        # filter sup and inf with supeq and infeq
        opidx_sup = self._filter_wrong_matches(opidx_supeq, opidx_sup)
        opidx_inf = self._filter_wrong_matches(opidx_infeq, opidx_inf)
        # split expr into values and operators
        # sorted operators so we can iterate over the expr from left to right
        opidx_all = sorted(opidx_eq + opidx_diff + opidx_supeq + opidx_infeq + opidx_sup + opidx_inf, key=lambda x: x[0])
        split_expr = []
        split_expr.append(expr[:opidx_all[0][0]])
        for i in range(len(opidx_all) - 1):
            # always take on the value on the right side of the op, so init the first part outside of the loop
            opidx_curr = opidx_all[i]
            opidx_next = opidx_all[i+1]
            operator = expr[opidx_curr[0]:opidx_curr[1]]
            split_expr.append(operator)
            value = expr[opidx_curr[1]:opidx_next[0]]
            split_expr.append(value)
        split_expr.append(expr[opidx_all[-1][0]:opidx_all[-1][1]])
        split_expr.append(expr[opidx_all[-1][1]:])
        return split_expr

    def _find_opidx(self, op, expr):
        """ Helper function for _split_expr.
        Return all occurrences indices of a comparison operator (op) within an expr.
        """
        # init possible operator symbols
        pattern = re.compile(op)  # raw string
        return [(m.start(0), m.end(0)) for m in re.finditer(pattern, expr)]

    def _filter_wrong_matches(self, opidx_larger, opidx_smaller):
        """Helper function for __split_expr.
        Filter out false positives of comparison operators. For instance,
        '<' beginning at the same position as '<=' should be discarded.
        """
        invalid = []
        for smaller in opidx_smaller:
            for larger in opidx_larger:
                if smaller[0] == larger[0]:
                    invalid.append(smaller)
        return [smaller for smaller in opidx_smaller if smaller not in invalid]



"""
Module fragment_combination_graph
=================================
This modules contains the functions for generating fragments combination graphs
from individual molecules:

It provides functions to annotate pseudo-Natural Products as well.
"""

# standard
import logging
import itertools
import toolz
# data handling
import pandas as pd
from pandas import DataFrame
from collections import OrderedDict
# chemoinformatics
from rdkit.Chem import Mol
# graph
import networkx as nx
# docs
from typing import List
# dev
from npfc import fragment_combination_point
from npfc import draw
from npfc import utils


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GLOBALS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

# for computing fragment graphs
DF_FG_COLS = ['idm',
              'inchikey',
              'idfcg',
              'nfrags',
              'nfrags_u',
              'ncomb',
              'ncomb_u',
              'hac_mol',
              'hac_frags',
              'perc_mol_cov_frags',
              '_frags',
              '_frags_u',
              '_comb',
              '_comb_u',
              'fcg_str',
              '_d_aidxs',
              '_colormap',
              '_fcg',
              'mol',
              '_d_mol_frags',
              '_d_fcp_labels',
              ]

# for annotating fragment graphs with PNP
DF_PNP_COLS = DF_FG_COLS + ['pnp_fm', 'pnp_mol']


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def _clear_ffs(df_fcc: DataFrame) -> DataFrame:
    """Clear ffs combinations by discarding any combination in which the smaller
    fragments is involved.

    :param df_fcc: a fcc DataFrame
    :return: a cleaned fcc DataFrame
    """

    # clean the data

    logging.debug("Now cleaning fragment combinations")

    # drop cutoff combinations
    logging.debug("Removing cutoff connections from fragment combinations")
    num_fcc_ini = len(df_fcc.index)
    logging.debug("Number of remaining fragment combinations: %s/%s", len(df_fcc.index), num_fcc_ini)

    # drop fragments combinations paired with a substructure
    logging.debug("Removing substructures from fragment combinations")
    df_substructures = df_fcc[df_fcc['fcc'] == 'ffs']  # all the substructures in the whole dataframe
    num_substructures = len(df_substructures.index)
    logging.debug("Number of substructures found in df_fcc: %s/%s", num_substructures, len(df_fcc.index))
    # in case of substructures to remove, iterate over all identified subtructures for each molecule,
    # determine what fragments are part of others and discard all entries with them
    if num_substructures > 0:
        logging.debug("Substructure combinations:\n\n%s\n", df_substructures[['idm', 'fc']])
        logging.debug("Determining what fragments should be removed:")
        # intialize the iteration
        rowids_to_remove = []  # the rowids of the df_fcc dataframe to remove
        for gid, g in df_fcc[df_fcc['idm'].isin(df_substructures['idm'])].groupby('idm'):  # iterate only on the groups with at least one substructure
            fid_to_remove = set()   # fid of substructures identified for the current molecule
            # for each molecule, look at what fids we should remove
            for rowid, row in g[g['fcc'] == 'ffs'].iterrows():
                # combination ifs ffs, so remove either fid1 or fid2 depending on hac
                if len(row['_aidxf1']) > len(row['_aidxf2']):
                    fid_to_remove.add(row['fid2'])
                else:
                    fid_to_remove.add(row['fid1'])
                # display some debugging
                logging.debug("%s: %s - %s ==> fid_to_remove: %s", gid, row['fid1'], row['fid2'], fid_to_remove)
                # register df_fcc rowids that will be removed for this substructure
                rowids_to_remove += list(g[g["fid1"].isin(list(fid_to_remove))].index) + list(g[g["fid2"].isin(list(fid_to_remove))].index)
        # remove dupl in rowids_to_remove
        rowids_to_remove = list(set(rowids_to_remove))
        # filter the unwanted fragment combinations
        logging.debug("Number of fragments combinations to remove: %s", len(rowids_to_remove))
        nb_fcc_ini = len(df_fcc.index)
        df_fcc = df_fcc.loc[~df_fcc.index.isin(rowids_to_remove)]
        logging.debug("Number of fragment combinations remaining: %s/%s", len(df_fcc), nb_fcc_ini)

    return df_fcc


def _get_incompatible_fragments_dict(df_overlaps: DataFrame) -> dict:
    """Compute a dictionary indicating what fragment id is incompatible with another:
    d[frag1] = [frag2, frag3, ...]
    d[frag2] = [frag1, frag3, ...]
    d[frag3] = [frag1, frag2, ...]

    :param df_overlaps: DataFrame with only fragment combinations of type ffo (overlaps.)
    :return: a dictionary for easy check of incompatibilities
    """
    d_incompatible = {}
    for i in range(len(df_overlaps.index)):
        row = df_overlaps.iloc[i]
        fid1 = row['fid1']
        fid2 = row['fid2']

        if fid1 not in d_incompatible.keys():
            d_incompatible[fid1] = [fid2]
        else:
            d_incompatible[fid1].append(fid2)

        if fid2 not in d_incompatible.keys():
            d_incompatible[fid2] = [fid1]
        else:
            d_incompatible[fid2].append(fid1)

    return d_incompatible


def _split_overlaps(df_fc: DataFrame, max_overlaps: int) -> DataFrame:
    """Split a fragment combination DataFrame containing overlap entries into different fragment combination DataFrames.

    This function is used within a loop, hence the need for gid parameter (logging).

    :param df_fc: fragment combination DataFrame
    :max_overlaps: maximum number of authorized overlap entries in the DataFrame, if observed number is higher, then an empty DataFrame is returned
    :return: a List of DataFrames
    """
    # Split df_fc into 3 dfs:
    # 1. overlaps: fc of type ffo
    # 2. commons: fc not involving any fragment in ffo
    # 3. variants: fc involving fragments in ffo
    # overlaps
    df_overlaps = df_fc[df_fc['fcc'] == 'ffo']
    noverlaps = len(df_overlaps.index)
    logging.debug("Number of overlaps found: %s'", noverlaps)

    if noverlaps == 0:
        return ([df_fc], noverlaps)
    elif noverlaps > max_overlaps:
        logging.debug("Too many overlap combinations (%s/%s), discarding molecule.", noverlaps, max_overlaps)
        return ([], noverlaps)

    # code below is for cases with at least 1 overlap
    d_incompatible = _get_incompatible_fragments_dict(df_overlaps)
    # common
    df_common = df_fc[(~df_fc['fid1'].isin(d_incompatible.keys())) & (~df_fc['fid2'].isin(d_incompatible.keys()))]
    logging.debug("Identified %s variant combinations:\n%s", len(df_common.index), df_common[['idm', 'fc']])

    # variants
    df_variants = df_fc[(df_fc['fcc'] != 'ffo') & ((df_fc['fid1'].isin(d_incompatible.keys()) | (df_fc['fid2'].isin(d_incompatible.keys()))))]
    logging.debug("Identified %s variant combinations:\n%s", len(df_variants.index), df_variants[['idm', 'fcc']])

    # define a list of sets of alternative fragments (i.e. [(A, B), (C, D)]) for computing all possible alternative fgraph possibilities
    alt_frags = list(set([tuple(sorted([k] + v)) for k, v in d_incompatible.items()]))
    # remove sublists included in others: i.e. ('1724:0', '627:1'), ('1724:0', '627:0'), ('1724:0', '627:0', '627:1') => ('1724:0', '627:0', '627:1')
    alt_frags = [set(x) for x in alt_frags]  # now the former set of tuples is a list of sets
    alt_frags.sort(key=len)  # do only 1 check: left in right and not right in left as well
    alt_frags = [tuple(sorted(list(x))) for x in list(filter(lambda f: not any(f < g for g in alt_frags), alt_frags))]  # discard any set that is subset of another
    logging.debug("Alternative fragments:\n" + '\n'.join([str(x) for x in alt_frags]))

    # compute 1 df for each alternative fgraph and then concatenate it all into one single df
    dfs_alt_curr = []
    for product in itertools.product(*alt_frags):
        logging.debug("Current alternative route for fgraphs: %s", product)
        to_remove = []
        for p in product:
            to_remove += d_incompatible[p]
        df_variants_curr = df_variants[((~df_variants['fid1'].isin(to_remove)) & (~df_variants['fid2'].isin(to_remove)))]
        dfs_alt_curr.append(pd.concat([df_common, df_variants_curr]))

    # unfortunately, there are a lot of duplicates going on after splitting by overlaps
    dfs_alt_curr = list(toolz.unique(dfs_alt_curr, key=lambda x: utils.encode_object(x)))

    # clear duplicate dfs
    return (dfs_alt_curr, noverlaps)


def _split_unconnected(dfs_fcc_clean: List[DataFrame]) -> List[DataFrame]:
    """
    From a list of DataFrames of Fragment Combinations, split up DataFrames containing
    two or more unconnected parts in different DataFrames.
    For instance if a molecule has two paired fragments left and two fragments right
    too far from other, then we get something like 2 and 2 instead of 4.

    :param dfs_fcc_clean: a List of fcc DataFrames
    :return: an updated List of connected fcc DataFrames
    """
    dfs_fcc_ready = []
    for i, df_fcc_clean in enumerate(dfs_fcc_clean):
        # compute a graph with each fid as a node, one row means an edge between 2 fid
        G = nx.from_pandas_edgelist(df_fcc_clean, "fid1", "fid2")  # we do not care for parallel edges here, just for the node ids in each subgraphs
        fc_subgraphs = list(G.subgraph(c) for c in nx.connected_components(G))
        num_fc_subgraphs = len(fc_subgraphs)
        # splitting up subgraphs
        if num_fc_subgraphs > 1:
            logging.debug("Fragment Connectivity -- %s: found %s subgraphs, so splitting up", i, num_fc_subgraphs)
            # for each subgraph, record corresponding rows in df only
            for fc_subgraph in fc_subgraphs:
                nodes = list(fc_subgraph.nodes())
                df_fcc_subgraph = df_fcc_clean[((df_fcc_clean['fid1'].isin(nodes)) | (df_fcc_clean['fid2'].isin(nodes)))].copy()
                dfs_fcc_ready.append(df_fcc_subgraph)
        else:
            dfs_fcc_ready.append(df_fcc_clean)

    return dfs_fcc_ready


def generate(df_fcc: DataFrame, min_frags: int = 2, max_frags: int = 5, max_overlaps: int = 5, split_unconnected: bool = True, clear_ffs: bool = True, palette: str = None) -> DataFrame:
    """This method process a fragment combinations DataFrame
    and return a new DataFrame with a fragment combination graph for each molecule.

    Each highlighted fragment of the molecule consists of a node of the graph.
    Fragment Combinations are used as edges for displaying fragment interactions and are annotated
    with the idm as well as the category of the combination.

    A str representation is also computed:
        frag1[cmo]frag2-frag2[fed]frag3

    Two objects are computed and stored as b64 strings:
        - colormap: a custom object with 3 informations regarding highlight colors (RGB values):
            - fragments: the color attributed to each fragment
            - atoms: the color attributed to each atom
            - bonds: the color attributed to each bond
        - graph: a nx object used comparing fragment connectivity among molecules.

    Molecules can be filtered using thresholds.

    :param df_fcc: a Dataframe with pairwise fragment combinations
    :param min_frags: a threshold for the minimum number of fragments allowed per fragment combination graph
    :param max_frags: a threshold for the maximum number of fragments allowed per fragment combination graph
    :param max_overlaps: a threshold for the maximum number of overlap combinations found in the molecule
    :return: a DataFrame representing fragment combination graphs
    """
    # split by overlaps

    logging.debug("Mapping fragments")

    use_fcp_labels = True
    ds_fcg = []
    for gid, g in df_fcc.groupby('idm'):
        logging.debug("Current Molecule: %s", gid)
        # split overlaps into different Dataframes
        dfs_fcc_clean, noverlaps = _split_overlaps(g, max_overlaps)

        if logging.getLogger().level == logging.DEBUG:
            for i, df_fcc_clean in enumerate(dfs_fcc_clean):
                logging.info("df_fcc_clean #%s\n\n%s\n", i, df_fcc_clean[['idm', 'fc']])

        if len(dfs_fcc_clean) == 0:
            continue

        # compute fragment connectivity graph objects so we can split up disconnected subgraphs
        if split_unconnected:
            dfs_fcc_ready = _split_unconnected(dfs_fcc_clean)
        else:
            dfs_fcc_ready = dfs_fcc_clean

        # clear ffs
        if clear_ffs:
            dfs_fcc_ready = [_clear_ffs(df_fcc_ready) for df_fcc_ready in dfs_fcc_ready]

        # compute the entries of the df_fcg
        for i, df_fcc_clean in enumerate(dfs_fcc_ready):

            # string representation of the fragment combinations of this map
            fragment_combination_graph_str = '-'.join(list(df_fcc_clean['fc'].map(str)))
            logging.debug('fcg_str: %s', fragment_combination_graph_str)

            # d_aidxs: a dict containing the occurrences of each fragment type
            d_aidxs = {}
            d_frags = {}
            d_fcp_labels = {}
            for j in range(len(df_fcc_clean.index)):
                row = df_fcc_clean.iloc[j]
                # idf1
                if row["idf1"] not in d_aidxs.keys():
                    d_aidxs[row["idf1"]] = [row["_aidxf1"]]
                    d_frags[row["idf1"]] = row["mol_frag_1"]
                    try:
                        d_fcp_labels[row['idf1']] = row['_fcp_labels_1']
                    except KeyError:
                        use_fcp_labels = False
                elif row["_aidxf1"] not in d_aidxs[row["idf1"]]:
                    d_aidxs[row["idf1"]].append(row["_aidxf1"])
                    try:
                        d_fcp_labels[row['idf1']] = row['_fcp_labels_1']
                    except KeyError:
                        use_fcp_labels = False
                # idf2
                if row["idf2"] not in d_aidxs.keys():
                    d_aidxs[row["idf2"]] = [row["_aidxf2"]]
                    d_frags[row["idf2"]] = row["mol_frag_2"]
                    try:
                        d_fcp_labels[row['idf2']] = row['_fcp_labels_2']
                    except KeyError:
                        use_fcp_labels = False
                elif row["_aidxf2"] not in d_aidxs[row["idf2"]]:
                    d_aidxs[row["idf2"]].append(row["_aidxf2"])
                    try:
                        d_fcp_labels[row['idf2']] = row['_fcp_labels_2']
                    except KeyError:
                        use_fcp_labels = False

            # sort d_aidxs for reproducible colormaps  ### might be the cause of the wrong coloration in alternative fgraphs due to overlaps
            d_aidxs = OrderedDict(sorted(d_aidxs.items()))
            if use_fcp_labels:
                d_fcp_labels = OrderedDict(sorted(d_fcp_labels.items()))
            else:
                logging.warning('FCP labels are not available')

            # count fragment occurrences (non-unique)
            frags = list(set([x for x in df_fcc_clean['fid1'].map(str).values] + [x for x in df_fcc_clean['fid2'].map(str).values]))
            nfrags = len(frags)

            # filter results by min/max number of fragment occurrences
            if nfrags < min_frags or nfrags > max_frags:
                logging.debug("%s: discarding one fragment combination graph because of unsuitable number of fragments (%s)", gid, nfrags)
                continue

            # count unique fragment types (unique)
            frags_u = list(d_aidxs.keys())
            nfrags_u = len(frags_u)

            # compute fragment coverage of the molecule
            hac_mol = g.iloc[0]['hac']  # same hac for all entries since this is the same molecule anyway
            # hac
            hac_frags = set()
            for k in d_aidxs.keys():
                hac_frags.update(set([item for sublist in d_aidxs[k] for item in sublist]))
            hac_frags = len(hac_frags)
            # perc
            perc_mol_cov_frags = round((hac_frags / hac_mol), 2) * 100

            # compute a new graph again but this time on a single subgraph and with edge labels (room for optimization)
            # count the number of equivalent edges (sames ids and same fcc)
            df_fcc_clean = df_fcc_clean.copy()  # ### one day I will have to understand why all of the Pandas warnings appear all the time
            # here I used to compress all combinations in common
            # df_fcc_clean['n_fcc'] = df_fcc_clean.groupby(['idf1', 'idf2', 'fcc'])['fcc'].transform('count')
            # df_fcc_clean.drop_duplicates(subset=["idf1", "idf2", "fcc"], keep="first", inplace=True)
            # create two new columns: fcp_1 and fcp_2 for indicating fragment connection points of source and target
            df_fcc_clean['fcp_1'], df_fcc_clean['fcp_2'] = zip(*df_fcc_clean['fc'].map(lambda x: (x.split('[')[0].split('@')[1], x.split(']')[1].split('@')[1])))
            # compute the graph
            edge_attr = ['fcc', 'n_fcc', 'idm', 'fcp_1', 'fcp_2']
            edge_attr = [x for x in edge_attr if x in df_fcc_clean.columns]
            df_fcc_clean = df_fcc_clean.sort_values(['idf1', 'idf2'])
            df_fcc_edges = df_fcc_clean.copy()[['idf1', 'idf2'] + edge_attr]
            G = nx.from_pandas_edgelist(df_fcc_edges, source="idf1", target="idf2", edge_attr=edge_attr, create_using=nx.MultiGraph())  # multigraph instead of grpah to store parallel edges
            # same molecule in each row, so to use the first one is perfectly fine
            mol = df_fcc_clean.iloc[0]['mol']

            # in case of overlaps, the same molecule will be used more than once,
            # so make a copy of the original so highlights are truly independant
            if noverlaps > 0:
                mol = Mol(mol)

            # attribute colors to each fragment atoms/bonds
            colormap = draw.ColorMap(mol, d_aidxs, palette)

            comb = list(df_fcc_clean['fcc'].values)
            ncomb = len(comb)
            comb_u = list(set(comb))
            ncomb_u = len(comb_u)
            if not use_fcp_labels:
                d_fcp_labels = {}

            ds_fcg.append({'idm': gid, 'inchikey': g.iloc[0]['inchikey'], 'idfcg': str(i+1).zfill(3), 'nfrags': nfrags, 'nfrags_u': nfrags_u, 'ncomb': ncomb, 'ncomb_u': ncomb_u, 'hac_mol': hac_mol, 'hac_frags': hac_frags, 'perc_mol_cov_frags': perc_mol_cov_frags, '_frags': frags, '_frags_u': frags_u, '_comb': comb, '_comb_u': comb_u, 'fcg_str': fragment_combination_graph_str, '_d_aidxs': d_aidxs, '_colormap': colormap, '_fcg': G, 'mol': mol, '_d_mol_frags': d_frags, '_d_fcp_labels': d_fcp_labels})

    # put it all together
    df_fcg = DataFrame(ds_fcg, columns=DF_FG_COLS).drop_duplicates(subset=['fcg_str'])
    df_fcg['idfcg'] = df_fcg.groupby('idm').cumcount().map(lambda x: str(x+1).zfill(3))
    # incorporate the idcfg to the graphs
    df_fcg.apply(lambda x: nx.classes.function.set_edge_attributes(x['_fcg'], x['idfcg'], 'idcfg'), axis=1)

    return df_fcg


def has_only_referenced_edges(edges: tuple, edges_ref: tuple) -> bool:
    """Check if at least one edge in edges is not present within edges_ref.
    Edges are tuple of syntax (u, v, d) with d being the dict with the attributes.

    :param edges: the edges of the target molecule fcg
    :param edges_ref: the edges of reference molecule fcg
    :return: False if at least 1 edge is not found in the reference, True otherwise
    """
    # hotfix: ####
    edges_ref = [tuple([set([er[0], er[1]]), er[2]]) for er in edges_ref]
    edges = [tuple([set([e[0], e[1]]), e[2]]) for e in edges]

    logging.debug('Reference edges: %s', edges_ref)
    for e in edges:
        logging.debug('current edge: %s', e)
        if e not in edges_ref:
            logging.debug('edge is not present within references!')
            return False
    return True


def filter_edges_attributes(edges: list, cols: list) -> list:
    """Networkx can either return one or all properties. Now that I am aware of this,
    I extract all attributes and then filter out the ones I do not want to use.

    :param edges: the edges as a list of tuple of syntax (u, v, d) with d being the dict with the attributes
    :param cols: the list of attributes to use for PNP labelling.
    """
    return [list([row[0], row[1], {k: v for k, v in row[2].items() if k in cols}]) for row in edges]


def get_pnp_references(edges: tuple, df_ref: DataFrame, target_node: frozenset = None) -> tuple:
    """Return a tuple of reference idms. A reference is recorded if all edges of the target molecule are included  at once
    within the reference row edges. In case no reference is found, an empty tuple is returned.

    :param edges: the edges as a list of tuple of syntax (u, v, d) with d being the dict with the attributes
    :param df_ref: the dataframe containing the edges to use for references.
    :param target_nodes: a frozenset of fragment ids found in the target edges to use for filering references to compare (optimization)
    """
    frags_u = frozenset([x[0] for x in edges] + [x[1] for x in edges])
    logging.debug('Input unique idfs to use: %s', frags_u)
    # avoid unnecessary graph comparisons: no way the syntehtic compound will be matching a NP fc graph if it does not even have all the nodes
    df_ref = df_ref[df_ref['_frags_u'].map(lambda x: frags_u.issubset(x))]
    logging.debug('Remaining reference entries after filtering by unique idfs: %s', len(df_ref))

    # get the subset of references that match with this particular molecule
    df_ref = df_ref[df_ref['edges'].map(lambda x: has_only_referenced_edges(edges, x))]

    # if no ref is found, just label as PNP
    if len(df_ref) < 1:
        return tuple()
    else:
        df_ref['idm_idfcg'] = df_ref['idm'].astype(str) + ':' + df_ref['idfcg'].astype(str)
        return tuple(df_ref['idm_idfcg'].values)


def annotate_pnp(df_fcg, df_fcg_ref, data=['fcc', 'fcp_1', 'fcp_2'], consider_symmetry=True) -> DataFrame:
    """Search and Identify for PNP molecules in the input DataFrame (df_fcg).
    PNP molecules are defined as molecules containing natural fragments combinations
    that are not found in a reference natural dataset.
    Fragment combinations are defined by extracting information from networkx graphs containing 3 informations:

        - source: id of fragment 1
        - target: id of fragment 2
        - attributes to consider: by default: fcc, fcp_1, fcp_2, etc.

    Three new columns are appended to the input DataFrame:

        - _pnp_ref: the list of a references found for the target fcg
        - pnp_fcg: True if the input fcg has no match with any reference fcg, False otherwise
        - pnp_mol: True if the input molecule has no matching fcg with any reference fcg, False otherwise

    When considering FCP, it is recommanded to consider symmetry. This results in ignoring the suffixes in FCPs,
    i.e. '1a' and '1b' become both '1'.

    :param df_fcg: the input DataFrame
    :param df_fcg_ref: the reference DataFrame
    :param data: the list of edge attributes to consider during fcg comparison
    :param consider_symmetry: consider fragment symmetry during annotating when using FCPs. To use FCPs data must include 'fcp_1' and/or 'fcp_2' (using or would not make sense here).
    :return: the input DataFrame with 3 new pnp columns
    """

    # extract edge informations for targets
    df_fcg["edges"] = df_fcg["_fcg"].map(lambda x: x.edges(data=True))
    df_fcg["edges"] = df_fcg["edges"].map(lambda x: filter_edges_attributes(x, data))
    df_fcg['_frags_u'] = df_fcg['_frags_u'].map(lambda x: frozenset(x))

    # extract edge informations for references
    df_fcg_ref["edges"] = df_fcg_ref["_fcg"].map(lambda x: x.edges(data=True))
    df_fcg_ref["edges"] = df_fcg_ref["edges"].map(lambda x: filter_edges_attributes(x, data))
    df_fcg_ref['_frags_u'] = df_fcg_ref['_frags_u'].map(lambda x: frozenset(x))

    # handle symmetry in case it is wished for
    if consider_symmetry and 'fcp_1' in data and 'fcp_2' in data:
        df_fcg['edges'] = df_fcg['edges'].map(fragment_combination_point.clear_fcp_suffixes_in_edges)
        df_fcg_ref['edges'] = df_fcg_ref['edges'].map(fragment_combination_point.clear_fcp_suffixes_in_edges)

    # run annotation
    df_fcg['_pnp_ref'] = df_fcg.apply(lambda x: get_pnp_references(x['edges'], df_fcg_ref, x['_frags_u']), axis=1)
    df_fcg['pnp_fcg'] = df_fcg['_pnp_ref'].map(lambda x: True if len(x) == 0 else False)
    df_fcg['pnp_mol'] = df_fcg.groupby('idm')[['pnp_fcg']].transform(lambda x: True if any(x) else False)

    # drop edges column since edges are included intact in G
    df_fcg = df_fcg.drop('edges', axis=1)

    return df_fcg


def regroup_edges_from_fcgs(df_fcg):
    """This function regroups all edges of a molecule. It can be applied to more than only one molecule at once.
    This will result in one single fcg per molecule, with all combinations and fragment occurrences counted only once,
    which is useful for reporting. (I should never have split fragment graphs).
    """
    idm_groups = df_fcg.groupby('idm')
    rows = []
    cols = ['idm', 'idfcg', 'idf1', 'fcp_1', 'fcc', 'idf2', 'fcp_2']
    for gid, g in idm_groups:
        combinations = set('-'.join(g['fcg_str']).split('-'))
        for c in combinations:
            f1 = c.split(':')[0]
            f2 = c.split(']')[1].split(':')[0]
            fcc = c.split('[')[1].split(']')[0]
            fcp1 = c.split('@')[1].split('[')[0]
            fcp2 = c.split('@')[-1]
            rows.append([gid, -1, f1, fcp1, fcc, f2, fcp2])
    df_edges = pd.DataFrame(rows, columns=cols)
    return df_edges


def get_ref_aidxs(df_fs):
    """Part of the hotfix for redundant FCGs.
    I did not record the occurrence id in the graphs, which was stupid.
    So now I need to use the df_fs to get the information instead.
    Needs to be used with fid col, which is defined in filter_out_fcgs_ffs_all.
    """
    return {k: v for k, v in zip(df_fs['fid'], df_fs['_aidxf'])}


def get_varying_d_aidxs(varying_fragments_occ, d):
    """Part of the hotfix for redundant FCGs.
    Attribute the atom indices to the fragmnet occurrences.
    """
    return {k: set(v) for k, v in d.items() if k in varying_fragments_occ}


def _keep_first_fcg(d1, d2):
    """Part of the hotfix for redundant FCGs.
    Determine if:

        0. both FCGs should be kept
        1. first FCG should be kept
        2. second FCG should be kept

    This is based on atom indices only.
    """
    c1 = 0
    c2 = 0
    # print(f"{d1=}")
    # print(f"{d2=}")
    for f1, a1 in d1.items():
        for f2, a2 in d2.items():
            # print(f"investigating: {f1} ({a1}) + {f2} ({a2})")
            if a1.issubset(a2):
                # print(f"{f1} included in {f2}")
                c1 += 1
            elif a2.issubset(a1):
                # print(f"{f2} included in {f1}")
                c2 += 1

    # print(f"{c1=}; {c2=}")

    if c1 == 0 and c2 > 0:
        return 1
    elif c2 == 0 and c1 > 0:
        return 2
    else:
        return 0


def filter_out_fcgs_ffs(df2, d):
    """Part of the hotfix for redundant FCGs.
    Apply on DF with FCGs of a same molecules.
    """
    df2 = df2.rename({'_frags': 'frags'}, axis=1)

    to_remove = []

    for i, row1 in enumerate(df2.itertuples()):
        frags1 = set(row1.frags)
        if row1.idfcg in to_remove:
            continue
        for row2 in df2.iloc[i+1:].itertuples():
            if row2.idfcg in to_remove:
                continue
            # print(f"comparing FCG {row1.idfcg} + FCG {row2.idfcg}")
            # print('\n' + "=" * 60 + '\n')
            # print(f"COMPARISON: FCG {row1.idfcg} + FCG {row2.idfcg}".center(60))
            # print('\n' + "=" * 60 + '\n')
            # define common / varying fragments
            frags2 = set(row2.frags)
            # print(f"{frags1=}")
            # print(f"{frags2=}")
            common_fragments_occ = frags1.intersection(frags2)
            varying_fragments_occ1 = frags1.difference(frags2)
            varying_fragments_occ2 = frags2.difference(frags1)

            # print("\ncommon_fragments_occ")
            # display(common_fragments_occ)
            # print("\nvarying_fragments_occ1")
            # display(varying_fragments_occ1)
            # print("\nvarying_fragments_occ2")
            # display(varying_fragments_occ2)

            # display(d)

            d1 = get_varying_d_aidxs(varying_fragments_occ1, d)
            d2 = get_varying_d_aidxs(varying_fragments_occ2, d)

            keep_first = _keep_first_fcg(d1, d2)
            if keep_first == 1:
                to_remove.append(row2.idfcg)
            elif keep_first == 2:
                to_remove.append(row1.idfcg)

            # print(f"keep_first: {keep_first}")
            # if keep_first == 1:
                # print(f"FCG {row2.idfcg} will be removed")
            # elif keep_first == 2:
                # print(f"FCG {row1.idfcg} will be removed")
            # else:
            #     pass
                # print("keep both FCG")

            # do not waste time on comparing other possibilities
            # when we already know that this FCG has to go
            if keep_first == 2:
                break

    # print('\n' + "=" * 60 + '\n')
    # print("RESULT".center(60))
    # print('\n' + "=" * 60 + '\n')

    to_remove = set(to_remove)
    # print("FCGs to remove:")
    # display(to_remove)

    return df2[~df2['idfcg'].isin(to_remove)].rename({'frags': '_frags'}, axis=1)


def filter_out_fcgs_ffs_all(df_fcg, df_fs):
    """Part of the hotfix for redundant FCGs.
    """
    dfs = []
    df_fs['fid'] = df_fs['idf'].astype(str) + ':' + df_fs['idf_idx'].astype(str)
    for gid, g in df_fcg.groupby('idm'):
        d_ref = get_ref_aidxs(df_fs[df_fs['idm'] == gid])
        dfs.append(filter_out_fcgs_ffs(g, d_ref))

    return pd.concat(dfs)



"""
Module utils
================
"""

# standard
import logging
from pathlib import Path
import sys
import itertools
import signal
from contextlib import contextmanager
import string
from random import choice
# data handling
import pickle
import base64
# typing
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import Mol
# chemoinformatics
from typing import Union
from typing import List


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GLOBALS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


# allowed suffixes
EXTS_INPUT = [['.sdf'], ['.sdf', '.gz'],
              ['.csv'], ['.csv', '.gz'],
              ['.log'],  # for loading formatted log files
              ['.hdf'],
              ['.pkl', '.gz'], ['.model', '.gz']  # model files for computing sa and np scores
              ]  # , ['.feather']]  # tests with feather work but not in production, dig into that later

EXTS_CONFIG = [['.json'], ['.mplstyle']]

# types
Number = Union[int, float]
Output_files = List[List[Union[str, int]]]


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def get_file_format(input_file: str) -> tuple:
    """Deduce how the file should be parsed based on its suffixes.

    Only compressions for one single file with initial extension still visibile
    work (see example below).

    >>> from pathlib import Path
    >>> from npfc import utils
    >>> utils.get_file_format(Path('file.csv.gz').suffixes)
    >>> # returns ('CSV', 'gzip')
    >>> utils.get_file_format(Path('file.sdf').suffixes)
    >>> # returns ('SDF', None)

    :param suffixes: suffixes of a file
    :return: a tuple with syntax (format, compression)
    """
    suffixes = Path(input_file).suffixes
    logging.debug("Input file Suffixes are: %s", suffixes)
    # is the file an archive?
    if len(suffixes) > 1:
        compression = suffixes[1]
        logging.debug("File is an archive with compression=%s", compression)
        # special case for gzip so .gz files can be read directly with pandas
        if compression == '.gz':
            compression = 'gzip'
        else:
            raise ValueError("Error! Unexpected value for compression suffix: '%s'.", compression)
    else:
        compression = None

    # return format and compression
    if suffixes[0] in ('.sdf', '.mol', '.sd'):
        return ('SDF', compression)
    elif suffixes[0] == '.csv':
        return ('CSV', compression)
    elif suffixes[0] in ('.hdf', 'hf5'):
        return ('HDF', compression)
    elif suffixes[0] == ".feather":
        return ('FEATHER', compression)


def check_arg_bool(value: bool) -> bool:
    """Return True of the value is indeed a boolean, raise a TypeError otherwise.

    :param value: the argument to test
    """
    if not isinstance(value, bool):
        raise TypeError(f"Error! Expected a boolean but got {type(value)} instead ({value}).")

    return True


def parse_argparse_boolstring(value: str) -> bool:
    """Return True or False given the provided string. If the string is actually not a boolean, raise a TypeError.

    :param value: the argument to test
    :return: the parsed boolean as type bool
    """
    if value.lower() in ("yes", "true", "t", "1"):
        return True
    elif value.lower() in ("no", "false", "f", "0"):
        return False
    else:
        raise TypeError(f"Error! Could not parse provided boolean string ({value}).")


def check_arg_positive_number(value: Number) -> bool:
    """Return True of the value is indeed a positive number (>0), raise a TypeError otherwise.

    :param value: the argument to test
    """
    logging.debug("value is %s (%s)", value, type(value))
    # possible value for Number is None if argument is left unset
    if value is None:
        return True
    # if not None, then looks what it is
    if not isinstance(value, Number.__args__):  # might be a hack but Union object is not compatible with isinstance
        raise TypeError(f"Error! Expected a positive number but got {type(value)} instead ({value}).")
    elif value <= 0:
        raise ValueError(f"Error! Expected a positive number but got {value} instead.")

    return True


def check_arg_input_file(input_file: str) -> bool:
    """Return True of the input_file exists, raise an error otherwise.

    :param input_file: the input file
    :param input_format: the expected format of the input file
    """
    path_input_file = Path(input_file)
    if not path_input_file.is_file():
        raise ValueError(f"Error! Input file could not be found at '{input_file}'.")
    if path_input_file.suffixes not in EXTS_INPUT:
        raise ValueError(f"Error! Unexpected '{path_input_file.suffixes}' for input format.")

    return True


def check_arg_input_dir(input_dir: str) -> bool:
    """Return True of the input_dir exists.

    :param input_dir: the output directory
    :return: True if the directory exists.
    """
    # output_format
    p = Path(input_dir)
    if not p.is_dir():
        raise ValueError(f"Error! Input dir could not be found at '{input_dir}'.")

    return True


def check_arg_output_file(output_file: str, create_parent_dir: bool = True) -> bool:
    """Return True of the output_file has the expected format (deduced from the file extension).

    If the parent directory of the output file does not exist, it has to either be created or fail the check.

    :param output_file: the output file
    :param create_parent_dir: create the output file's parent folder in case it does not exist
    """
    # output_format
    path_output_file = Path(output_file)
    if path_output_file.suffixes not in EXTS_INPUT:
        raise ValueError(f"Error! Unexpected value '{path_output_file.suffixes}' for output format.")

    # create_parent_dir
    output_dir = path_output_file.resolve().parent
    if not output_dir.is_dir():
        if create_parent_dir:
            logging.warning(f"Output_dir could not be found at '{output_dir}', attempting to create it.")
            output_dir.mkdir(parents=True, exist_ok=True)
        else:
            raise ValueError(f"Error! Output_dir could not be found at '{output_dir}'.")

    return True


def check_arg_output_plot(output_file: str, create_parent_dir: bool = True) -> bool:
    """Return True of the output_plot has the expected format (deduced from the file extension).

    Accepted extensions are: svg and png.

    If the parent directory of the output file does not exist, it has to either be created or fail the check.

    :param output_file: the output file
    :param create_parent_dir: create the output file's parent folder in case it does not exist
    """
    # output_format
    path_output_file = Path(output_file)
    if path_output_file.suffixes not in [['.svg'], ['.png']]:
        raise ValueError(f"Error! Unexpected value '{path_output_file.suffixes}' for output plot.")

    # create_parent_dir
    output_dir = path_output_file.resolve().parent
    if not output_dir.is_dir():
        if create_parent_dir:
            logging.warning(f"Output_dir could not be found at '{output_dir}', attempting to create it.")
            output_dir.mkdir(parents=True, exist_ok=True)
        else:
            raise ValueError(f"Error! Output_dir could not be found at '{output_dir}'.")

    return True


def check_arg_output_dir(output_dir: str) -> bool:
    """Return True of the output_dir can exist.

    If the parent directory of the output dir does not exist, it has to either be created or fail the check.

    :param output_dir: the output directory
    :param create_parent_dir: create the output file's parent folder in case it does not exist
    """
    # output_format
    path_output_file = Path(output_dir)
    if not path_output_file.is_dir():
        path_output_file.mkdir(parents=True, exist_ok=True)

    return True


def check_arg_input_config(config_file: str) -> bool:
    """Return True of the config_file exists, raise an error otherwise.

    :param input_file: the input file
    """
    if config_file is not None:
        path_config_file = Path(config_file)
        if not path_config_file.is_file():
            raise ValueError(f"Error! Input file could not be found at '{config_file}'.")
        elif path_config_file.suffixes not in EXTS_CONFIG:
            raise ValueError(f"Error! Unexpected '{path_config_file.suffixes}' for config format.")

    return True



def check_arg_output_config(output_config: str, create_parent_dir: bool = True) -> bool:
    """Return True of the output_config has the expected format (deduced from the file extension).

    If the parent directory of the output file does not exist, it has to either be created or fail the check.

    :param output_config: the output config file
    :param create_parent_dir: create the output file's parent folder in case it does not exist
    """
    # output_format
    path_output_config = Path(output_config)
    if path_output_config.suffixes not in EXTS_CONFIG:
        raise ValueError(f"Error! Unexpected value '{path_output_config.suffixes}' for output format.")

    # create_parent_dir
    output_dir = path_output_config.resolve().parent
    if not output_dir.is_dir():
        if create_parent_dir:
            logging.warning(f"Output_dir could not be found at '{output_dir}', attempting to create it.")
            output_dir.mkdir(parents=True, exist_ok=True)
        else:
            raise ValueError(f"Error! Output_dir could not be found at '{output_dir}'.")

    return True

def _configure_logger(log_level: str, logger_name: str = None, log_file: str = None, reset_handlers: bool = True) -> logging:
    """Configure the logging in a centralized way. This is useful for scripts mostly.

    :param log_level: the logging level to use, accepted values are: CRITICAL, ERROR, WARNING, INFO, DEBUG.
    :param logger_name: specify a name for the logger. If none, __name__ is used. If within a loop, better speficy a logger name, as logging messages would otherwise stack.
    :param log_file: if specified, the logging messages will be written to the corresponding file, in addition to stdout.
    :return: a logger object that can be reset
    """
    if reset_handlers:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
    if logger_name is None:
        logger_name = __name__
    # define level with DEBUG instead of logging.DEBUG or numbers
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f'Invalid log level: {log_level}')

    # create the logger
    logger = logging.getLogger(logger_name)
    if len(logger.handlers) > 0:
        logger.handlers.clear()
    logger.setLevel(numeric_level)
    # define format
    if log_level == 'DEBUG':
        format_string = ("%(asctime)s -- %(levelname)s -- %(funcName)s:%(lineno)d -- %(message)s")
    else:
        format_string = ("%(asctime)s -- %(levelname)s -- %(message)s")
    log_format = logging.Formatter(format_string)

    # Creating and adding the console handler
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setFormatter(log_format)
    logger.addHandler(console_handler)

    # Creating and adding the file handler
    file_handler = logging.FileHandler(logger_name, mode='w')
    file_handler.setFormatter(log_format)
    logger.addHandler(file_handler)

    # avoid tons of duplicated log messages when executing code from modules
    logger.propagate = False

    return logger


def encode_object(element: object) -> str:
    """Convert an object to a base64 string.

    :param element: an object to encode
    :return: an base64 string upon success, None otherwise
    """
    try:
        return base64.b64encode(pickle.dumps(element)).decode("utf-8")
    except TypeError:
        return None


def decode_object(string: str) -> object:
    """Convert a base64 string to an object.

    :param string: a base64 string encoding an object
    :return: an object upon success, None otherwise
    """
    try:
        return pickle.loads(base64.b64decode(string))
    except TypeError:
        return None


def encode_mol(mol: Mol) -> str:
    """Convert a molecule to a base64 string.

    :param mol: the input molecule
    :return: the molecule in base64
    """
    try:
        return base64.b64encode(mol.ToBinary()).decode("utf-8")
    except AttributeError:
        return None


def decode_mol(string: str) -> Mol:
    """Convert a string to a RDKit Mol object.

    :param string: a string with a Mol object in bytes with a base64 string representation
    :return: a Mol object upon success, None otherwise

    """
    try:
        return Mol(base64.b64decode(string))
    except TypeError:
        return None


def encode_mol_smiles(mol: Mol) -> str:
    """Convert a mol to a Smiles.

    :param string: a string with a Mol object in bytes with a base64 string representation
    :return: a str object upon success, None otherwise

    """
    try:
        return Chem.MolToSmiles(mol)
    except TypeError:
        return None


def decode_mol_smiles(string: str) -> Mol:
    """Convert a Smiles to a Mol.

    :param string: a string with a Mol object in bytes with a base64 string representation
    :return: a Mol object upon success, None otherwise

    """
    if string is None:
        string = ''
    try:
        return Chem.MolFromSmiles(string)
    except TypeError:
        return None


def get_shortest_path_between_frags(mol: Mol, aidxf1: set, aidxf2: set) -> tuple:
    """Return the shortest path within a molecule between two fragments defined by atom indices.
    First and last atom indices are part of respectively fragment 1 and fragment 2, so they should not
    be considered when estimating the distance between fragments.

    (i.e. distance = len(shortest_path) - 2)

    :param mol: The input molecule.
    :param aidxf1: the atom indices of the first fragment found in the molecule
    :param aidxf2: the atom indices of the second fragment found in the molecule
    :return: the atom indices of the shortest path between both fragments. The first index is the attachment point from fragment 1 whereas the last index is the attachment point from fragment 2
    """
    # 1/ compute every pairwise atom combination between both fragments
    pairwise_combinations = itertools.product(tuple(aidxf1), tuple(aidxf2))
    # 2/ for each of those, compute the shortest path possible
    pairwise_combinations = list(pairwise_combinations)
    all_paths = [Chem.GetShortestPath(mol, pc[0], pc[1]) for pc in pairwise_combinations]
    logging.debug("Looking for the shortest path shortest path among these:")
    [logging.debug("Path (%s): %s", str(i).zfill(3), p) for i, p in enumerate(all_paths)]
    # 3/ return one of the shortest pathes
    return min(all_paths, key=lambda x: len(x))


def fuse_rings(rings: tuple) -> list:
    """
    Check for atom indices in common between rings to aggregate them into fused rings.

    :param rings: the ring atoms as provided by the RDKit function mol.GetRingInfo().AtomRings() (iteratble of iteratable of atom indices).
    :return: the fused ring atoms (list of lists of atom indices)
    """
    # condition to exit from recursive fusion of ring atom indices
    done = False

    while not done:
        fused_rings = []
        num_rings = len(rings)
        # pairwise check for common atoms between rings
        for i in range(num_rings):
            # define a core
            fused_ring = set(rings[i])
            for j in range(i+1, num_rings):
                # detect if ring is fused
                if set(rings[i]) & set(rings[j]):
                    # add fused ring to our rign atom list
                    fused_ring = fused_ring.union(rings[j])
            # either lone or fused ring, first check if not already in fused_rings
            if any([fused_ring.issubset(x) for x in fused_rings]):
                continue
            fused_rings.append(list(fused_ring))
        rings = list(fused_rings)
        # there are no rings to fuse anymore
        if num_rings == len(rings):
            done = True

    return rings


@contextmanager
def timeout(time):
    """This function is used within a with statement:

    >>> with timeout(5):
    >>>    do something

    If the code block execution time exceeds the time threshold, a TimeoutError is raised.

    :param time: time in seconds allowd to the code block before aborting its execution

    References

        - https://www.jujens.eu/posts/en/2018/Jun/02/python-timeout-function/
        - https://docs.python.org/3/library/contextlib.html

    """
    # register a function to raise a TimeoutError on the signal.
    signal.signal(signal.SIGALRM, raise_timeout)
    # schedule the signal to be sent after time
    signal.alarm(time)
    # run the code block within the with statement
    try:
        yield
    except TimeoutError:
        pass  # exit the with statement
    finally:
        # unregister the signal so it won't be triggered if the timeout is not reached
        signal.signal(signal.SIGALRM, signal.SIG_IGN)


def raise_timeout(signum, frame):
    """Function to actually raise the TimeoutError when the time has come.
    """
    raise TimeoutError


def random_string(n: int) -> str:
    """Function to generate a random stringand of n characters and digits and
    in lower case.

    :param n: the number of characters in the generated string
    :return: a random string
    """
    return ''.join([choice(string.ascii_lowercase + string.digits) for i in range(n)])



"""
Module fragment_search
==========================
This modules contains the function to run substructure searches.
"""


# standard
import logging
# data handling
from pandas import DataFrame
# chemoinformatics
from rdkit.Chem import rdTautomerQuery


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FUNCTIONS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def get_fragment_hits(df_mols: DataFrame,
                      df_frags: DataFrame,
                      col_mol_mols: str = 'mol',
                      col_mol_frags: str = 'mol',
                      col_mol_inchikey: str = 'inchikey',
                      fcp_labels: str = None,
                      tautomer: bool = False,
                      col_to_index_mols: str = 'idm',
                      col_to_index_frags: str = 'idm',
                      ) -> DataFrame:
    """Create a DataFrame recording every Fragment Hit in the
    input molecule DataFrame.

    A Fragment Hit is composed of 6 fields:

    1) idm: the id of the molecule (rowid from df_mols)
    2) idf: the id of the fragment (rowid from df_frags)
    3) aidxf: the atom indices of the fragment found in the molecule
    4) mol_perc: the percentage of the molecule the fragment represents (based on hac)
    5) mol: the molecule as RDKit Mol object
    6) mol_frag: the fragment as RDKit Mol object

    :param df_mols: the input DataFrame with the molecules (df_mols)
    :param df_frags: the input DataFrame with fragments to use for substructure search (df_frags)
    :param col_mol_mols: the column name in df_mols with the molecules
    :param col_mol_frags: the column name in df_frags with the fragments
    :param col_mol_inchikey: the input DataFrame column name with the inchikey of the molecule. If it does not exist, then an empty column is created in the output.
    :param fcp_labels: the column name in the fragments dataframe with the fcp labels
    :param tautomer: if set to True, tautomers will be taken into account during fragment search (warning, tautomer-independant search is much slower!)
    :param col_to_index_mols: set the row indices of the DataFrame with the molecules to probe to the specified column. If empty (''), indices are left untouched.
    :param col_to_index_frags: set the row indices of the DataFrame with the fragments to seach for to the specified column. If empty (''), indices are left untouched.

    :return: the substructure matches as a DataFrame

    .. warning:: Row indices are used for recording the ids of substructure hits and are therefore required to be set to the molecule identifiers (i.e. idm).

    """
    # init
    d = {}
    d['idm'] = []
    d['idf'] = []
    d['_aidxf'] = []
    d['mol_perc'] = []  # proportion of the molecule the substructure represents
    d['mol'] = []  # encode the molecule here so we don't have to combine multiple files when trying to have a look at the results
    d['mol_frag'] = []  # strucutre of the fragment
    d['inchikey'] = []  # inchikey of the molecule

    # row ids are used for storing the correct molecule and fragment ids in the
    #  output table, so they need to be set before the computation occurs
    if col_to_index_mols != '':
        df_mols.index = list(df_mols[col_to_index_mols])
    if col_to_index_frags != '':
        df_frags.index = list(df_frags[col_to_index_frags])

    # if no inchikey was computed, generate an empty column
    if col_mol_inchikey not in df_mols.columns:
        df_mols['inchikey'] = ''

    # tautomers
    if tautomer:
        df_mols[col_mol_mols + '_taut'] = df_mols[col_mol_mols].map(lambda x: rdTautomerQuery.TautomerQuery(x).GetTemplateMolecule())

    # begin
    for i in range(len(df_mols.index)):
        rowm = df_mols.iloc[i]
        if tautomer:
            mol = rowm[col_mol_mols + '_taut']
        else:
            mol = rowm[col_mol_mols]
        hac = mol.GetNumAtoms()
        for j in range(len(df_frags.index)):
            rowq = df_frags.iloc[j]
            # perform the substructure search on mol so the latest matching fragment does not get highlighted
            matches = mol.GetSubstructMatches(rowq[col_mol_frags])
            if len(matches) > 0:
                logging.debug("MOL %s + FRAG %s ==> %s", rowm.name, rowq.name, matches)
                for m in matches:
                    d['idm'].append(rowm.name)
                    d['idf'].append(rowq.name)
                    d['_aidxf'].append(m)  # frozenset so we can use intersection, etc. and still remove dupl. easily
                    d['mol_perc'].append(round(len(m)/hac, 2) * 100)
                    d['mol'].append(rowm[col_mol_mols])
                    d['inchikey'].append(rowm[col_mol_inchikey])
                    d['mol_frag'].append(rowq[col_mol_frags])

    df_fs = DataFrame(d)
    df_fs['idf_idx'] = df_fs.groupby(['idm', 'idf']).cumcount()  # rank seems to be working with np.float types only...

    # add fcp labels
    if fcp_labels is not None and fcp_labels != '':
        df_frags = df_frags.rename({'idm': 'idf'}, axis=1)
        df_fs = df_fs.merge(df_frags[['idf', fcp_labels]], on='idf', how='left')

    return df_fs.rename({fcp_labels: '_fcp_labels'}, axis=1)



"""
Module standardize
===================
This modules is used to standardize molecules and molecular DataFrames.
"""

# standard
import logging
from collections import Counter
from copy import deepcopy
from more_itertools import intersperse
import pkg_resources
# data handling
import json
import pandas as pd
from pandas import DataFrame
# chemoinformatics
from rdkit.Chem import AllChem as Chem
from rdkit.Chem import Mol
from rdkit.Chem import rdinchi
from rdkit.Chem import Descriptors
from rdkit.Chem import rdmolops
from rdkit.Chem.MolStandardize.metal import MetalDisconnector
from rdkit.Chem.MolStandardize.charge import Uncharger
from rdkit.Chem.MolStandardize.normalize import Normalizer
from rdkit.Chem.MolStandardize.tautomer import TautomerCanonicalizer
from rdkit.Chem.Scaffolds import MurckoScaffold
# graph
from networkx import Graph
# docs
from typing import Union
from typing import List
# dev library
from npfc.draw import depict_mol
from npfc import utils
from npfc.filter import Filter


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GLOBALS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

DEFAULT_ELEMENTS = {'H', 'B', 'C', 'N', 'O', 'F', 'P', 'S', 'Cl', 'Br', 'I'}

# DEFAULT_PROTOCOL = {'tasks': ['filter_empty',
#                               'disconnect_metal',
#                               'clear_mixtures',
#                               'deglycosylate',
#                               'filter_num_heavy_atom',
#                               'filter_molecular_weight',
#                               'filter_num_ring',
#                               'filter_elements',
#                               'clear_isotopes',
#                               'normalize',
#                               'uncharge',
#                               'canonicalize',
#                               'clear_stereo',
#                               ],
#                     'filter_num_heavy_atom': 'num_heavy_atom > 3',
#                     'filter_molecular_weight': 'molecular_weight <= 1000.0',
#                     'filter_num_ring': 'num_ring > 0',
#                     'filter_elements': f'elements in {", ".join(str(x) for x in DEFAULT_ELEMENTS)}',
#                     'filter_unwanted': ['c1ccccc1'],
#                     }


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASSES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


class FullUncharger(Uncharger):
    """A class derived from rdkit.Chem.MolStandardize.charge.Uncharger, so
    instead of attempting to create zwitterions all possible charges are removed
    from the molecule.

    For instance:

    >>> # Uncharger:
    >>> [O-][N+](C)(C)C[O-] -> [O-][N+](C)(C)CO

    >>> # FullUncharger:
    >>> [O-][N+](C)(C)C[O-] -> O[N+](C)(C)CO
    """

    def __init__(self):
        """Create an instance of FullUncharger.

        .. todo:: This will remove charges from -2 to +2 only. This could be improved using more general smarts?
        """
        # some smarts to use to find charges
        self.q_pos_1 = Chem.MolFromSmarts("[*;+]")
        self.q_pos_2 = Chem.MolFromSmarts("[*;+2]")
        self.q_neg_1 = Chem.MolFromSmarts("[*;-]")
        self.q_neg_2 = Chem.MolFromSmarts("[*;-2]")
        logging.debug("Initialized a new FullUncharger object")

    def full_uncharge(self, mol: Mol) -> Mol:
        """Neutralize molecule by adding/removing hydrogens.
        Does not attempt to preserve zwitterions.
        For now takes into account only charges of -2 and +2.

        :param mol: the input molecule
        :return: the uncharged molecule
        """
        logging.debug("Uncharging a molecule")
        mol = deepcopy(mol)
        # Get atom ids for matches
        p = [x[0] for x in mol.GetSubstructMatches(self.q_pos_1)]   # +1
        p += [x[0] for x in mol.GetSubstructMatches(self.q_pos_2)]  # +2
        n = [x[0] for x in mol.GetSubstructMatches(self.q_neg_1)]  # -1
        n += [x[0] for x in mol.GetSubstructMatches(self.q_neg_2)]  # -2
        # remove positive charges
        for atom in [mol.GetAtomWithIdx(x) for x in p]:
            # Remove hydrogen and reduce formal change until neutral or no more hydrogens
            while atom.GetFormalCharge() > 0 and atom.GetNumExplicitHs() > 0:
                atom.SetNumExplicitHs(atom.GetNumExplicitHs() - 1)
                atom.SetFormalCharge(atom.GetFormalCharge() - 1)
        # remove negative charges
        for atom in [mol.GetAtomWithIdx(x) for x in n]:
            # Add hydrogen and increase formal change until neutral
            while atom.GetFormalCharge() < 0:
                atom.SetNumExplicitHs(atom.GetNumExplicitHs() + 1)
                atom.SetFormalCharge(atom.GetFormalCharge() + 1)

        # clean-up
        mol.ClearComputedProps()
        mol.UpdatePropertyCache()
        Chem.GetSymmSSSR(mol)

        # mol.ClearComputedProps()  # not testes but might solved the -O-H2 issue
        # mol.UpdatePropertyCache()
        return mol


class Standardizer(Filter):
    """A class for standardizing molecular structures. The standardization itself is based
    on a protocol that the user can modify.

    By default this protocol consists in 15 tasks applied to each molecule invidually:

        1) **initiate_mol**: check if the molecule passed the RDKit conversion
        2) **filter_empty**: filter molecules with empty structures
        3) **disconnect_metal**: break bonds involving metallic atoms, resulting in potentially several molecules per structure.
        4) **clear_mixtures**: retrieve only the "best" molecule from a mixture, which might not always be the largest one.
        5) **deglycosylate**: remove all external sugars-like rings from the molecule and return the remaining non-linear entity.
        6) **filter_num_heavy_atom**: filter molecules with a heavy atom count not in the accepted range. By default: num_heavy_atom > 3.
        7) **filter_molecular_weight**: filter molecules with a molecular weight not in the accepted range. By default: molecular_weight <= 1000.0.
        8) **filter_num_ring**: filter molecules with a number of rings (Smallest Sets of Smallest Rings or SSSR) not in the accepted range. By default: num_ring > 0.
        9) **filter_elements**: filter molecules with elements not considered as medchem. By default: elements in H, B, C, N, O, F, P, S, Cl, Br, I.
        10) **clear_isotopes**: set all atoms to their most common isotope (i.e. 14C becomes 12C which is C).
        11) **normalize**: always write the same functional groups in the same manner.
        12) **uncharge**: remove all charges on a molecule when it is possible. This is different from rdkit.Chem.MolStandardize.charge module as there is no attempt for reaching the zwitterion.
        13) **canonicalize**: enumerate the canonical tautomer.
        14) **clear_stereo**: remove all remaining stereochemistry flags on the molecule.
        15) **reset_mol**: convert forth and back to SMILES format to discard potential residual outdated flags on atoms and bonds.

    Other steps are not part of this protocol but can be executed as well for convenience:

        - **depict**: find the "best" possible 2D depiction of the molecule among Input/rdDepictor/Avalon/CoordGen methods
        - **extract_murcko**: return the Murcko Scaffold from the molecule
        - **clear_side_chains**: remove any exocyclic atom that is not part of a linker
        - **reset_mol**: reset the molecule by converting to and then from smiles

    This results in new columns in the input DataFrame:

        - the 'mol' column: updated structure (only for the protocol)
        - the 'status' column: either passed, filtered or error.
        - the 'task' column: the latest task that was applied to the molecule.

    The standardizer works either on a molecule (method: 'run') or on a DataFrame containing molecules ('run_df').

    In the latter case, the inchikey is computed and can be used for identifying duplicate entries.

    A timeout value is set by default and will be applied to each molecule individually to avoid the process being stuck on marginally difficult cases.
    This value can be set either during the Standardizer object initialization or by defining as an option in the protocol (priority is given to the latter if defined).
    """

    def __init__(self,
                 protocol: str = None,
                 col_mol: str = 'mol',
                 col_id: str = 'idm',
                 elements_medchem: set = DEFAULT_ELEMENTS,
                 timeout: int = 10,
                 ):
        """Create a Standardizer object.

        :param protocol: Either a JSON file or a dictionary. The resultung dictinary needs a 'tasks' key that lists all tasks to be excuted as a list.
        :param col_mol: the column with the molecule for when running the run_df method
        :param col_id: the column with the id for when running the run_df method
        :param filter_duplicates:
        """
        # filter
        super(Standardizer, self).__init__()
        # standardizer
        self._elements_medchem = elements_medchem
        self._col_id = col_id
        self._col_mol = col_mol

        if protocol is None:
            self._protocol = json.load(open(pkg_resources.resource_filename('npfc', 'data/std_mols.json'), 'r'))
        else:
            if isinstance(protocol, str):
                self._protocol = json.load(open(protocol, 'r'))
            else:
                self._protocol = protocol
        # workers
        self.metal_disconnector = MetalDisconnector()
        self.normalizer = Normalizer()
        self.full_uncharger = FullUncharger()
        self.canonicalizer = TautomerCanonicalizer()

        # display information on protocol
        if logging.getLogger().level == logging.DEBUG:
            logging.debug("Successfully instanciated a Standardizer object with protocol:")
            [logging.debug("Task #%s: %s", str(i+1).zfill(2), task) for i, task in enumerate(self._protocol['tasks'])]
            [logging.debug("Option %s %s", opt, value) for opt, value in self._protocol.items() if opt != 'tasks']

    def __repr__(self):
        return f"Standardizer ({len(self._protocol['tasks'])} tasks)"

    def describe(self):
        # init
        pad = max(len(x) for x in self._protocol['tasks'])
        head = 'STANDARIDZER={\n'
        tail = '\n}'
        # define a list of tasks with options in parenthesis
        tasks = list(self._protocol['tasks'])
        for i, task in enumerate(tasks):
            if task in self._protocol.keys():
                opt = self._protocol[task].replace(task.replace('filter_', ''), 'x')
                tasks[i] = f"{task} ({opt})"
        # concatenate all parts and intersperse the tasks with bottow arrows, with step index on the left
        return head + '\n'.join(intersperse('↓'.center(pad+10), [str(i+1).zfill(2).ljust(5) + x.center(pad) for i, x in enumerate(tasks)])) + tail

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, protocol: str):
        # input is a json file => convert it to a dict
        if isinstance(protocol, str):
            utils.check_arg_input_config(protocol)
            with open(protocol) as f:
                protocol = json.load(f)
        # input is a dict
        if isinstance(protocol, dict):
            if 'tasks' not in protocol.keys():
                raise ValueError("invalid protocol format (no 'tasks' key found)")
            elif not isinstance(protocol['tasks'], list) and not isinstance(protocol['tasks'], tuple):
                raise ValueError("invalid protocol format ('tasks' key is neither list or tuple)")
        # update default protocol
        self._protocol.update(protocol)

    @property
    def timeout(self) -> str:
        return self._timeout

    @timeout.setter
    def timeout(self, value: int) -> None:
        if not isinstance(value, int):
            raise ValueError(f"Error! timeout should be a positive int (>1), not '{type(value)}'.")
        elif value < 1:
            raise ValueError(f"Error! timeout should be superior to 1 ({value})")
        self._col_id = value

    @property
    def col_id(self) -> str:
        return self._col_id

    @col_id.setter
    def col_id(self, value: str) -> None:
        if value is None:
            raise ValueError(f"Error! col_id cannot be '{value}'.")
        self._col_id = value

    @property
    def col_mol(self) -> str:
        return self._col_mol

    @col_mol.setter
    def col_mol(self, value: str) -> None:
        if value is None:
            raise ValueError(f"Error! col_mol cannot be '{value}'.")
        self._col_mol = value

    @property
    def elements_medchem(self) -> set:
        return self._elements_medchem

    @elements_medchem.setter
    def elements_medchem(self, value: set) -> None:
        if not isinstance(value, set):
            raise ValueError(f"Error! elements_medchem should be a set of strings, not '{value}' ({type(value)}).")
        elif not all([isinstance(v, str) for v in value]):
            raise ValueError(f"Error! elements_medchem should be a set of strings, not '{value}' ({type(value)}).")
        self._elements_medchem = value

    def clear_isotopes(self, mol: Mol) -> Mol:
        """Return a molecule without any isotopes.

        :param mol: the input molecule
        :return: the molecule without isotope
        """
        mol = Mol(mol)
        for a in mol.GetAtoms():
            a.SetIsotope(0)
        return mol

    def clear_mixtures(self, mol: Mol) -> Mol:
        """Return the "best" molecule found in a molecular structure.

        The "best" molecule is determined by the following criteria, sorted by priority:

            1) contains only medchem elements
            2) contains at least one ring
            3) has the largest molecular weight of the mixture

        To summarize:

        .. math::
            medchem > non linear > molecular weight

        So the largest molecule of a mixture might not always be selected, for instance
        a very long aliphatic chain would be dismissed to keep a benzene molecule instead.

        This is implemented in such a way because our fragments used for substructure search contain at least one ring.
        On the contrary, this long aliphatic chain would be kept in a mixture with a non-medchem molecule.

        :param mol: the input molecule(s)
        :return: the best molecule
        """
        submols = Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=True)
        # no need to look further if we have only one submol!
        if len(submols) < 2:
            return mol
        # otherwise, we have to compare the submols
        # init
        logging.debug("found %s submols", len(submols))
        best_molecular_weight = -1.0  # so we are sure to update this on the first iteration
        best_submol = None
        best_is_medchem = False
        best_is_non_linear = False
        # begin
        for i, submol in enumerate(submols):
            # is_medchem
            is_medchem = self.filter_mol(submol, f'elements in {", ".join(str(x) for x in self.elements_medchem)}')
            is_non_linear = self.filter_mol(submol, "num_rings > 0")
            # molecular_weight
            molecular_weight = Descriptors.ExactMolWt(submol)
            logging.debug("submol #%s: IM=%s, INL=%s, MW=%s", i, is_medchem, is_non_linear, molecular_weight)
            # compare to the current best fragment
            update_best = False
            compute_diff = False  # check which
            # 2 criteria more important than molecular weight: is_medchem > is_non_linear
            if not best_is_medchem and is_medchem:
                update_best = True
            elif best_is_medchem and not is_medchem:
                continue
            elif not best_is_medchem and not is_medchem:
                if not best_is_non_linear and is_non_linear:
                    update_best = True
                elif best_is_non_linear and not is_non_linear:
                    continue
                else:
                    compute_diff = True
            else:  # best_is_medchem and is_medchem
                if not best_is_non_linear and is_non_linear:
                    update_best = True
                elif best_is_non_linear and not is_non_linear:
                    continue
                else:
                    compute_diff = True

            # check molecular_weights only in case of doubt
            if not update_best and compute_diff and molecular_weight > best_molecular_weight:
                update_best = True

            # update best with the properties of the current mol
            if update_best:
                best_is_medchem = is_medchem
                best_is_non_linear = is_non_linear
                best_submol = submol
                best_molecular_weight = molecular_weight

        return best_submol

    def clear_side_chains(self, mol: Mol, debug: bool = False) -> Mol:
        """Clear the side chains of a molecule.

        This method operates in 3 steps:

            1. Remove quickly all atoms in side chains but the one attached to a ring, starting from the terminal atom. (would certainly fail in case of linear molecules)
            2. Iterate over each remaining exocyclic atoms to remove only atoms when it does not break the ring aromaticity. Simple and double bonds can be broken and the atoms in rings which were attached to removed atoms are neutralized.
            3. Remove eventual nitrogen radicals by Smiles editing.

        .. warning:: I found only nitrogen radicals in my dataset, this might be insufficient on a larger scale.

        .. warning:: I found a bug for this molecule 'O=C(O)C1OC(OCC2OC(O)C(O)C(O)C2O)C(O)C(O)C1O', where a methyl remains after processing.

        :param mol: the molecule to simplify
        :return: a simplified copy of the molecule
        """
        # 1st peeling: fast, chunks of terminal chains
        smarts = Chem.MolFromSmarts('[!#1;R0;D1]~[!#1;R0;D{1-2}]')  # terminal exocyclic atom linked to another exocyclic atom, neighbour atom is not allowed more than 2 degrees, so branches (i.e. CC(=O)C) are not cut out
        while mol.HasSubstructMatch(smarts):
            mol = Chem.DeleteSubstructs(mol, smarts)
            mol.ClearComputedProps()
            mol.UpdatePropertyCache()
            Chem.GetSymmSSSR(mol)
        # 2nd peeling: slow, atom per atom of the remaining termninal atoms
        rwmol = Chem.RWMol(mol)
        smarts = Chem.MolFromSmarts('[!#1;R0;D1]')  # remaining terminal exocyclic atoms require cautious handling
        matches = sorted([item for sublist in rwmol.GetSubstructMatches(smarts) for item in sublist], reverse=True)  # reverse order so that remaining atom indices from matches are still valid after removing an atom
        for m in matches:
            try:
                m = m[0]  # should be only single atoms
                rwmol_tmp = deepcopy(rwmol)
                neighbor = rwmol_tmp.GetAtomWithIdx(m).GetNeighbors()[0]  # terminal atom so only 1 neighbor
                rwmol_tmp.RemoveAtom(m)
                neighbor.SetFormalCharge(0)  # neutralize in case of previously quaternary nitrogens
                neighbor.SetNumRadicalElectrons(0)  # remove radicals,this does not work as expected
                Chem.SanitizeMol(rwmol_tmp)  # will fail in case of break in aromaticity
                rwmol = rwmol_tmp  # if it went ok
            except Chem.rdchem.KekulizeException:
                pass  # we should not have tried to remove this atom, so just leave it be

        # I could not figure out how to remove radicals, so I just convert the mol to Smiles and edit the string
        return Chem.MolFromSmiles(Chem.MolToSmiles(rwmol).replace('[N]', 'N').replace('[n]', 'n'))

    def _is_sugar_like(self, ring_aidx: list, mol: Mol):
        """Indicate whether a ring (defined by its atom indices) in a molecule is sugar-like or not.

        Several conditions are to be met for a ring to be considered sugar-like:

            1. size: either 5 or 6 atoms
            2. elements: 1 oxygen and the rest carbons
            3. hybridization: ring atoms need have single bonds only
            4. connection points (next to the ring oxygen): at least 1 has an oxygen as neighbor
            5. subsituents (not next tot the ring oxygen): at least 1/2 (for 5-6-membered rings) have an oxygen as neighbor

        :param ring_aidx: the molecule indices of the ring to investigate
        :param mol: the molecule that contain the ring
        :return: True if the ring complies to the 5 conditions above, False otherwise.
        """
        # ring size: only 5-6 membered rings, rings are already fused when this function is called
        ring_size = len(ring_aidx)
        if ring_size != 5 and ring_size != 6:
            return False

        # access the actual atom objects quickier
        ring_atoms = [mol.GetAtomWithIdx(x) for x in ring_aidx]  # ring atoms are in the same order as ring_aidx

        # atom composition
        elements = [x.GetAtomicNum() for x in ring_atoms]
        element_counter = Counter(elements)
        if not ((ring_size == 5 and element_counter[6] == 4 and element_counter[8] == 1) or (ring_size == 6 and element_counter[6] == 5 and element_counter[8] == 1)):
            return False

        # hybridization of carbon atoms (check if only single bonds attached to the ring)
        carbon_atoms = [x for x in ring_atoms if x.GetAtomicNum() == 6]
        if any([x for x in carbon_atoms if x.GetHybridization() != 4]):  # to check if no H attached in case of the * position
            return False

        # to define connection points and substituents, we first need to identify the position of the ring oxygen
        oxygen_aidx = [x for x in ring_atoms if x not in carbon_atoms][0].GetIdx()  # only 1 oxygen in ring

        # connection points: 1 need at least 1 oxygen as neighbor
        cps = []
        cps_ok = False
        for carbon_atom in carbon_atoms:
            neighbors = carbon_atom.GetNeighbors()
            # if the ring oxygen is next to this atom, this atom is a connection point
            if any([n.GetIdx() == oxygen_aidx for n in neighbors]):
                cps.append(carbon_atom)
                # at least 1 of the connection points has to have an oxygen as side chain
                if any([n.GetAtomicNum() == 8 and n.GetIdx() != oxygen_aidx for n in neighbors]):
                    cps_ok = True
        if not cps_ok:
            return False

        # substituents
        substituents = [x for x in carbon_atoms if x.GetIdx() not in [y.GetIdx() for y in cps]]
        count_oxygens = 0
        for substituent in substituents:
            side_chain_atoms = [x for x in substituent.GetNeighbors() if x.GetIdx() not in ring_aidx]
            if len(side_chain_atoms) > 0:
                if not side_chain_atoms[0].GetAtomicNum() == 8:  # do not check for the degree here because there are connections on substituents too!
                    return False
                count_oxygens += 1
        # at least 1 oxygen for 5-membered rigns and 2 for 6-membered rings
        if (ring_size == 6 and count_oxygens < 2) or (ring_size == 5 and count_oxygens < 1):
            return False

        return True

    def deglycosylate(self, mol: Mol, mode: str = 'run') -> Union[Mol, Graph]:
        """Function to deglycosylate molecules.

        Several rules are applied for removing Sugar-Like Rings (SLRs) from molecules:

            1. Only external SLRs are removed, so a molecule with aglycan-SLR-aglycan is not modified
            2. Only molecules with both aglycans and SLRs are modified (so only SLRs or none are left untouched)
            3. Linear aglycans are considered to be part of linkers and are thus never returned as results
            4. Glycosidic bonds are defined as either O or CO and can be linked to larger linear linker. So from a SLR side, either nothing or only 1 C are allowed before the glycosidic bond oxygen
            5. Linker atoms until the glycosidic bond oxygen atom are appended to the definition of the SLR, so that any extra methyl is also removed.


        .. image:: _images/std_deglyco_algo.svg
            :align: center

        :param mol: the input molecule
        :param mode: either 'run' for actually deglycosylating the molecule or 'graph' for returning a graph of rings instead (useful for presentations or debugging)
        :return: the deglycosylated molecule or a graph of rings
        """

        if len(Chem.GetMolFrags(mol)) > 1:
            raise ValueError("Error! Deglycosylation is designed to work on single molecules, not mixtures!")

        if mode not in ('run', 'graph'):
            raise AttributeError(f"Error! Unauthorized value for parameter 'mode'! ('{mode}')")

        # avoid inplace modifications
        mol = Chem.Mol(mol)

        # define rings
        rings = mol.GetRingInfo().AtomRings()
        rings = utils.fuse_rings(rings)
        # try to deglycosylate only if the molecule has at least 2 rings:
        # - leave linear compounds out
        # - leave sugars in case they are the only ring on the molecule
        if len(rings) < 2:
            return mol

        # annotate sugar-like rings
        are_sugar_like = [self._is_sugar_like(x, mol) for x in rings]
        logging.debug('RINGS: %s', [(rings[i], are_sugar_like[i]) for i in range(len(rings))])
        # remove sugars only when the molecule has some sugar rings and is not entirely composed of sugars
        if not any(are_sugar_like) or all(are_sugar_like):
            return mol
        ring_atoms = set([item for sublist in rings for item in sublist])

        # init sugar graph
        G = Graph()
        # init linkers parts from left and right the glycosidic bond oxygen: one of the side is required to have either C or nothing
        authorized_linker_parts = [[], ['C']]  # R1-OxxxxR2 or R1-COxxxxR2 with xxxx being any sequence of linear atoms (same for R2->R1)

        # define linker atoms as shortest path between 2 rings that do not include other rings
        for i in range(len(rings)):
            ring1 = rings[i]
            for j in range(i+1, len(rings)):
                ring2 = rings[j]
                logging.debug('NEW RING PAIR -- R1: %s; R2: %s', ring1, ring2)

                # shortest path between the two rings that do not include the current rings themselves
                shortest_path = [x for x in Chem.GetShortestPath(mol, ring1[0], ring2[0]) if x not in ring1 + ring2]
                # define the other ring atoms
                other_ring_atoms = ring_atoms.symmetric_difference(set(ring1 + ring2))
                # shortest path for going from the left (ring1) to the right (ring2)
                shortest_path_elements = [mol.GetAtomWithIdx(x).GetSymbol() for x in shortest_path]

                # in case ring1 (left) or/and ring2 (right) is sugar-like, append the side chains left and right
                # to the oxygen to the corresponding ring atoms to avoid left-overs (the O remains is not removed)
                glycosidic_bond = False
                if 'O' in shortest_path_elements:  # not expected to be common enough for a try/catch statement
                    # from the left side
                    aidx_oxygen_left = shortest_path_elements.index('O')  # first O found in list
                    logging.debug('R1 -> R2 -- pos of O: %s; R1 is sugar_like: %s; linker part from R1: %s', aidx_oxygen_left, are_sugar_like[i], shortest_path_elements[:aidx_oxygen_left])
                    if are_sugar_like[i] and shortest_path_elements[:aidx_oxygen_left] in authorized_linker_parts:
                        glycosidic_bond = True
                        ring1 += shortest_path[:aidx_oxygen_left]

                    # from the right side
                    shortest_path_elements.reverse()
                    shortest_path.reverse()
                    aidx_oxygen_right = shortest_path_elements.index('O')  # first O found in list
                    logging.debug('R2 -> R1 -- pos of O: %s; R2 is sugar_like: %s; linker part from R2: %s', aidx_oxygen_right, are_sugar_like[j], shortest_path_elements[:aidx_oxygen_right])
                    if are_sugar_like[j] and shortest_path_elements[:aidx_oxygen_right] in authorized_linker_parts:
                        glycosidic_bond = True
                        ring2 += shortest_path[:aidx_oxygen_right]
                logging.debug('R1 and R2 are linked through a glycosidic bond: %s', glycosidic_bond)

                # in case the 2 rings are directly connected, append a new edge to G
                if not set(shortest_path).intersection(other_ring_atoms):
                    G.add_edge(i, j, atoms=''.join(shortest_path_elements), glycosidic_bond=glycosidic_bond)
                    # annotate nodes with the ring atoms (+ relevent linker atoms) and if they are sugar-like
                    G.nodes[i]['atoms'] = ring1
                    G.nodes[i]['sugar_like'] = are_sugar_like[i]
                    G.nodes[j]['atoms'] = ring2
                    G.nodes[j]['sugar_like'] = are_sugar_like[j]

        # draw the graph
        if mode == 'graph':
            # colormap_nodes = [(0.7,0.7,0.7) if x['sugar_like'] else (1,0,0) for i, x in G.nodes(data=True)]
            # return draw.fc_graph(G, colormap_nodes=colormap_nodes)
            return G

        # iterative recording of terminal sugar rings (atoms) that are linked with a glycosidic bond
        ring_atoms_to_remove = []
        nodes_to_remove = [node for node in G.nodes(data=True) if node[1]['sugar_like'] and G.degree(node[0]) == 1 and list(G.edges(node[0], data=True))[0][2]['glycosidic_bond']]
        while len(nodes_to_remove) > 0:
            # record atoms indices to remove from the molecule
            [ring_atoms_to_remove.append(n[1]['atoms']) for n in nodes_to_remove]
            # remove nodes from current layer for next iteration
            [G.remove_node(n[0]) for n in nodes_to_remove]
            nodes_to_remove = [node for node in G.nodes(data=True) if node[1]['sugar_like'] and G.degree(node[0]) == 1 and list(G.edges(node[0], data=True))[0][2]['glycosidic_bond']]
        logging.debug('Ring atoms to remove: %s', ring_atoms_to_remove)

        # edit the molecule
        if ring_atoms_to_remove:
            # flatten the atom indices of each ring to remove in reverse order so that atom indices do not change when removing atoms
            ring_atoms_to_remove = sorted([item for sublist in ring_atoms_to_remove for item in sublist], reverse=True)
            emol = Chem.EditableMol(mol)
            [emol.RemoveAtom(x) for x in ring_atoms_to_remove]
            mol = emol.GetMol()
            logging.debug('Obtained fragments: %s', Chem.MolToSmiles(mol))

        # clean-up
        frags = Chem.GetMolFrags(mol, asMols=True)
        # avoid counting the number of rings in each fragment if only 1 fragment left anyway
        if len(frags) == 1:
            logging.debug('Only one fragment obtained, returning it')
            return frags[0]
        # the substituents of the deleted terminal sugar-like rings remain in the structure,
        # these are obligatory linear because they were not in the graph,
        # so we just have to retrieve the one fragment that is not linear
        logging.debug('Returning only the non-linear obtained fragment')
        return [x for x in frags if Descriptors.rdMolDescriptors.CalcNumRings(x) > 0][0]

    def _run(self, mol: Mol) -> tuple:
        """Helper function for run.
        Contains all tasks defined within the protocol. Since some operations are
        expansive and could last a very long time for complex molecules (normalize, canonicalize),
        a timeout value is set globally. The run function is the one that can catch the exception
        raised by timeouts.

        :param mol: the input molecule
        :return: a tuple containing the molecule, its status and the further task name it reached
        """

        # initiate_mol
        if mol is None:
            return (mol, 'error', 'initiate_mol')

        # begin protocol
        mol = deepcopy(mol)  # do not modify the molecule in place
        for task in self._protocol['tasks']:
            # filter_empty
            if task == 'filter_empty':
                try:
                    if not mol.GetNumAtoms():
                        return ('', 'filtered', task)
                except ValueError:
                    return ('', 'error', task)

            # disconnect_metal
            elif task == 'disconnect_metal':
                try:
                    mol = self.metal_disconnector.disconnect(mol)
                except ValueError:
                    return (mol, 'error', 'disconnect_metal')

            # clear_mixtures
            elif task == 'clear_mixtures':
                try:
                    mol = self.clear_mixtures(mol)
                except ValueError:
                    return (mol, 'error', task)

            # deglycosylate
            elif task == 'deglycosylate':
                try:
                    mol = self.deglycosylate(mol)
                except ValueError:
                    return (mol, 'error', task)

            # filters
            elif task.startswith('filter_'):  # filter empty is tried before
                try:
                    if not self.filter_mol(mol, self._protocol[task]):
                        return (mol, 'filtered', task)
                except ValueError:
                    return (mol, 'error', task)

            # sanitize
            elif task == 'sanitize':
                try:
                    Chem.SanitizeMol(mol)
                except ValueError:
                    return (mol, 'error', task)

            # clear_isotopes
            elif task == 'clear_isotopes':
                try:
                    mol = self.clear_isotopes(mol)
                except ValueError:
                    return (mol, 'error', task)

            # normalize
            elif task == 'normalize':
                try:
                    mol = self.normalizer.normalize(mol)
                except ValueError:
                    return (mol, 'error', task)

            # uncharge
            elif task == 'uncharge':
                try:
                    mol = self.full_uncharger.full_uncharge(mol)
                except ValueError:
                    return (mol, 'error', task)

            # canonicalize
            elif task == 'canonicalize':
                # canonicalize
                try:
                    mol = self.canonicalizer.canonicalize(mol)
                except (ValueError, RuntimeError):
                    return (mol, 'error', task)

            # clear_stereo
            elif task == 'clear_stereo':
                try:
                    rdmolops.RemoveStereochemistry(mol)
                except ValueError:
                    return (mol, 'error', task)

            # extract Murcko Scaffolds
            elif task == 'extract_murcko':
                try:
                    mol = MurckoScaffold.GetScaffoldForMol(mol)
                except ValueError:
                    return (mol, 'error', task)

            # clear side chains
            elif task == 'clear_side_chains':
                try:
                    mol = self.clear_side_chains(mol)
                except ValueError:
                    return (mol, 'error', task)

            elif task == 'depict':
                try:
                    mol = depict_mol(mol)
                except ValueError:
                    return (mol, 'error', task)

            elif task == 'reset_mol':
                try:
                    mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
                except ValueError:
                    return (mol, 'error', task)

            # something else?
            else:
                raise ValueError(f"Unknown task: {task}")

        # a molecule that passed all the protocole!
        return (mol, 'passed', 'standardize')

    def run(self, mol: Mol, timeout: int = 10) -> tuple:
        """Execute the standardization protocol on a molecule.
        Molecule that exceed the timeout value are filtered with a task='timeout'.

        As a final step of the protocol, InChiKeys ('inchikey') are computed for identifying molecules.

        :param mol: the input molecule
        :param timeout: the maximum number of seconds for processing a molecule
        :return: a tuple containing the molecule, its status and the further task name it reached
        """
        with utils.timeout(timeout):
            return self._run(mol)

        # in case of timeout
        return (mol, 'filtered', 'timeout')


    def run_df(self, df: DataFrame, filter_unwanted: List[str] = []) -> tuple:
        """Apply the standardization protocol on a DataFrame, with the possibility of directly filtering duplicate entries as well.
        This can be very useful as the standardization process can expose duplicate entries due to salts removal, neutralization,
        canonical tautomer enumeration, and stereochemistry centers unlabelling

        If a reference file is specified, duplicate removals becomes possible accross chunks.


        :param df: the input DataFrame
        :param timeout: the maximum number of seconds for processing a molecule
        :param filter_unwanted: a list of molecules in SMILES format that should be filtered out at the end of the standardization
        :return: three DataFrames separated by status:

            - passed
            - filtered
            - error

        .. note:: As a side effect, the output DataFrames get indexed by idm.

        :param df: The DataFrame with molecules to standardize
        :param return: a tuple of 3 DataFrames: standardized, filtered and error.
        """
        df = df.copy()  # do not modify df in place
        # check filter_unwanted data (if any error, crash before running the protocol)
        if len(filter_unwanted) > 0:
            unwanted_molecules = [Chem.MolFromSmiles(x) for x in filter_unwanted]
            num_errors = 0
            for i, m in enumerate(unwanted_molecules):
                if m is None:
                    logging.error('Error in unwanted structure #%s! (%s)', i, filter_unwanted[i])
                    num_errors += 1
            if num_errors > 0:
                raise ValueError('Some errors were found in the unwanted molecules, aborting standardization!')
            unwanted_inchikeys = [rdinchi.MolToInchiKey(x) for x in unwanted_molecules]
            logging.debug('Unwanted Structure List:\n\n%s\n', DataFrame({'smiles': filter_unwanted, 'inchikey': unwanted_inchikeys}))

        # run standardization protocol
        df.index = df[self.col_id]
        df.loc[:, self.col_mol], df.loc[:, 'status'], df.loc[:, 'task'] = zip(*df[self.col_mol].map(self.run))

        # flag eventual None molecules at the end of the pipeline for filtering out
        df.loc[:, 'status'] = df.apply(lambda x: x['status'] if x['mol'] is not None else 'error', axis=1)
        df.loc[:, 'task'] = df.apply(lambda x: x['task'] if x['mol'] is not None else 'filter_empty_final', axis=1)
        df.loc[:, 'mol'] = df['mol'].map(lambda x: x if x is not None else '')

        # do not apply filter duplicates on molecules with errors or that were already filtered for x reasons
        df_error = df[df['status'] == 'error']
        df_filtered = df[df['status'] == 'filtered']
        df = df[df['status'].str.contains('passed')]

        # compute InChiKeys
        logging.debug('Computing InChI Keys for standardized structures')
        df.loc[:, 'inchikey'] = df.loc[:, self.col_mol].map(rdinchi.MolToInchiKey)

        # filter unwanted compounds using InChI Keys
        if len(filter_unwanted) > 0:
            logging.debug('Filtered out unwanted structures using InChI Keys')
            # identify unwanted compounds
            df_unwanted = df[df['inchikey'].isin(unwanted_inchikeys)]
            if len(df_unwanted) > 0:
                # remove these from passed compounds
                df = df[~df['idm'].isin(df_unwanted['idm'])]
                # annotate unwanted and add them to filtered compounds
                df.loc[:, 'task'] = 'standardize'
                df.loc[:, 'status'] = 'filter_unwanted'
                df_filtered = pd.concat([df_filtered, df_unwanted])
                logging.debug('Number of structures filtered out by identity: %s', len(df_unwanted.index))

        # tuple of dataframes
        return (df, df_filtered, df_error)
